#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""karbor common internal object model"""

import contextlib
import datetime

from oslo_log import log as logging
from oslo_versionedobjects import base
from oslo_versionedobjects import fields

from karbor import db
from karbor.db.sqlalchemy import models
from karbor import exception
from karbor.i18n import _
from karbor import objects


LOG = logging.getLogger('object')
remotable = base.remotable
remotable_classmethod = base.remotable_classmethod
obj_make_list = base.obj_make_list


class KarborObjectRegistry(base.VersionedObjectRegistry):
    def registration_hook(self, cls, index):
        setattr(objects, cls.obj_name(), cls)
        # For Versioned Object Classes that have a model store the model in
        # a Class attribute named model
        try:
            model_name = cls.obj_name()
            cls.model = getattr(models, model_name)
        except (ImportError, AttributeError):
            pass


class KarborObject(base.VersionedObject):
    OBJ_SERIAL_NAMESPACE = 'karbor_object'
    OBJ_PROJECT_NAMESPACE = 'karbor'

    def karbor_obj_get_changes(self):
        """Returns a dict of changed fields with tz unaware datetimes.

        Any timezone aware datetime field will be converted to UTC timezone
        and returned as timezone unaware datetime.

        This will allow us to pass these fields directly to a db update
        method as they can't have timezone information.
        """
        # Get dirtied/changed fields
        changes = self.obj_get_changes()

        # Look for datetime objects that contain timezone information
        for k, v in changes.items():
            if isinstance(v, datetime.datetime) and v.tzinfo:
                # Remove timezone information and adjust the time according to
                # the timezone information's offset.
                changes[k] = v.replace(tzinfo=None) - v.utcoffset()

        # Return modified dict
        return changes

    @base.remotable_classmethod
    def get_by_id(cls, context, id, *args, **kwargs):
        # To get by id we need to have a model and for the model to
        # have an id field
        if 'id' not in cls.fields:
            msg = (_('VersionedObject %s cannot retrieve object by id.') %
                   (cls.obj_name()))
            raise NotImplementedError(msg)

        model = getattr(models, cls.obj_name())
        orm_obj = db.get_by_id(context, model, id, *args, **kwargs)
        kargs = {}
        if hasattr(cls, 'DEFAULT_EXPECTED_ATTR'):
            kargs = {'expected_attrs': getattr(cls, 'DEFAULT_EXPECTED_ATTR')}
        return cls._from_db_object(context, cls(context), orm_obj, **kargs)

    def refresh(self):
        # To refresh we need to have a model and for the model to have an id
        # field
        if 'id' not in self.fields:
            msg = (_('VersionedObject %s cannot retrieve object by id.') %
                   (self.obj_name()))
            raise NotImplementedError(msg)

        current = self.get_by_id(self._context, self.id)

        for field in self.fields:
            # Only update attributes that are already set.  We do not want to
            # unexpectedly trigger a lazy-load.
            if self.obj_attr_is_set(field):
                if self[field] != current[field]:
                    self[field] = current[field]
        self.obj_reset_changes()

    def __contains__(self, name):
        # We're using obj_extra_fields to provide aliases for some fields while
        # in transition period. This override is to make these aliases pass
        # "'foo' in obj" tests.
        return name in self.obj_extra_fields or super(KarborObject,
                                                      self).__contains__(name)


class KarborObjectDictCompat(base.VersionedObjectDictCompat):
    """Mix-in to provide dictionary key access compat.

    If an object needs to support attribute access using
    dictionary items instead of object attributes, inherit
    from this class. This should only be used as a temporary
    measure until all callers are converted to use modern
    attribute access.

    NOTE(berrange) This class will eventually be deleted.
    """

    def get(self, key, value=base._NotSpecifiedSentinel):
        """For backwards-compatibility with dict-based objects.

        NOTE(danms): May be removed in the future.
        """
        if key not in self.obj_fields:
            # NOTE(jdg): There are a number of places where we rely on the
            # old dictionary version and do a get(xxx, None).
            # The following preserves that compatibility but in
            # the future we'll remove this shim altogether so don't
            # rely on it.
            LOG.debug('Karbor object %(object_name)s has no '
                      'attribute named: %(attribute_name)s',
                      {'object_name': self.__class__.__name__,
                       'attribute_name': key})
            return None
        if (value != base._NotSpecifiedSentinel and
                not self.obj_attr_is_set(key)):
            return value
        else:
            try:
                return getattr(self, key)
            except (exception.ObjectActionError, NotImplementedError):
                # Exception when haven't set a value for non-lazy
                # loadable attribute, but to mimic typical dict 'get'
                # behavior we should still return None
                return None


def DateTimeField(**kwargs):
    return fields.DateTimeField(tzinfo_aware=False, **kwargs)


class KarborPersistentObject(object):
    """Mixin class for Persistent objects.

    This adds the fields that we use in common for all persistent objects.
    """
    fields = {
        'created_at': DateTimeField(nullable=True),
        'updated_at': DateTimeField(nullable=True),
        'deleted_at': DateTimeField(nullable=True),
        'deleted': fields.BooleanField(default=False),
    }

    @contextlib.contextmanager
    def obj_as_admin(self):
        """Context manager to make an object call as an admin.

        This temporarily modifies the context embedded in an object to
        be elevated() and restores it after the call completes. Example
        usage:

           with obj.obj_as_admin():
               obj.save()
        """
        if self._context is None:
            raise exception.OrphanedObjectError(method='obj_as_admin',
                                                objtype=self.obj_name())

        original_context = self._context
        self._context = self._context.elevated()
        try:
            yield
        finally:
            self._context = original_context


class KarborComparableObject(base.ComparableVersionedObject):
    def __eq__(self, obj):
        if hasattr(obj, 'obj_to_primitive'):
            return self.obj_to_primitive() == obj.obj_to_primitive()
        return False


class ObjectListBase(base.ObjectListBase):
    pass


class KarborObjectSerializer(base.VersionedObjectSerializer):
    OBJ_BASE_CLASS = KarborObject


class DictOfDictOfStringsField(fields.AutoTypedField):
    AUTO_TYPE = fields.Dict(fields.Dict(fields.String(), nullable=True))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import abc
from collections import namedtuple

from oslo_log import log as logging
from oslo_serialization import jsonutils

import six

from karbor import exception
from karbor.i18n import _
from karbor.resource import Resource


_GraphBuilderContext = namedtuple("_GraphBuilderContext", (
    "source_set",
    "encountered_set",
    "finished_nodes",
    "get_child_nodes",
))

GraphNode = namedtuple("GraphNode", (
    "value",
    "child_nodes",
))

PackedGraph = namedtuple('PackedGraph', ['nodes', 'adjacency'])

LOG = logging.getLogger(__name__)


class FoundLoopError(RuntimeError):
    def __init__(self):
        super(FoundLoopError, self).__init__(
            _("A loop was found in the graph"))


def _build_graph_rec(context, node):
    LOG.trace("Entered node: %s", node)
    source_set = context.source_set
    encountered_set = context.encountered_set
    finished_nodes = context.finished_nodes
    LOG.trace("Gray set is %s", encountered_set)
    if node in encountered_set:
        raise FoundLoopError()

    LOG.trace("Black set is %s", finished_nodes.keys())
    if node in finished_nodes:
        return finished_nodes[node]

    LOG.trace("Change to gray: %s", node)
    encountered_set.add(node)
    child_nodes = context.get_child_nodes(node)
    LOG.trace("Child nodes are %s", child_nodes)
    # If we found a parent than this is not a source
    source_set.difference_update(child_nodes)
    child_list = []
    for child_node in child_nodes:
        child_list.append(_build_graph_rec(context, child_node))

    LOG.trace("Change to black: %s", node)
    encountered_set.discard(node)
    graph_node = GraphNode(value=node, child_nodes=tuple(child_list))
    finished_nodes[node] = graph_node

    return graph_node


def build_graph(start_nodes, get_child_nodes_func):
    context = _GraphBuilderContext(
        source_set=set(start_nodes),
        encountered_set=set(),
        finished_nodes={},
        get_child_nodes=get_child_nodes_func,
    )

    result = []
    for node in start_nodes:
        result.append(_build_graph_rec(context, node))

    assert(len(context.encountered_set) == 0)

    return [item for item in result if item.value in context.source_set]


@six.add_metaclass(abc.ABCMeta)
class GraphWalkerListener(object):
    """Interface for listening to GraphWaler events

    Classes that want to be able to use the graph walker to iterate over
    a graph should implement this interface.
    """
    @abc.abstractmethod
    def on_node_enter(self, node, already_visited):
        pass

    @abc.abstractmethod
    def on_node_exit(self, node):
        pass


class GraphWalker(object):
    def __init__(self):
        super(GraphWalker, self).__init__()
        self._listeners = []

    def register_listener(self, graph_walker_listener):
        self._listeners.append(graph_walker_listener)

    def unregister_listener(self, graph_walker_listener):
        self._listeners.remove(graph_walker_listener)

    def walk_graph(self, source_nodes):
        self._walk_graph(source_nodes, set())

    def _walk_graph(self, source_nodes, visited_nodes):
        for node in source_nodes:
            for listener in self._listeners:
                listener.on_node_enter(node, node in visited_nodes)
                visited_nodes.add(node)

            self._walk_graph(node.child_nodes, visited_nodes)

            for listener in self._listeners:
                listener.on_node_exit(node)


class PackGraphWalker(GraphWalkerListener):
    """Pack a list of GraphNode

    Allocate a serialized id (sid) for every node and build an adjacency list,
    suitable for graph unpacking.
    """
    def __init__(self, adjacency_list, nodes_dict):
        super(PackGraphWalker, self).__init__()
        self._sid_counter = 0
        self._node_to_sid = {}
        self._adjacency_list = adjacency_list
        self._sid_to_node = nodes_dict

    def on_node_enter(self, node, already_visited):
        pass

    def on_node_exit(self, node):
        def key_serialize(key):
            return hex(key)

        if node not in self._node_to_sid:
            node_sid = self._sid_counter
            self._sid_counter += 1
            self._node_to_sid[node] = node_sid
            self._sid_to_node[key_serialize(node_sid)] = node.value

            if len(node.child_nodes) > 0:
                children_sids = map(lambda node:
                                    key_serialize(self._node_to_sid[node]),
                                    node.child_nodes)
                self._adjacency_list.append(
                    (key_serialize(node_sid), tuple(children_sids))
                )


def pack_graph(start_nodes):
    """Return a PackedGraph from a list of GraphNodes

    Packs a graph into a flat PackedGraph (nodes dictionary, adjacency list).
    """
    walker = GraphWalker()
    nodes_dict = {}
    adjacency_list = []
    packer = PackGraphWalker(adjacency_list, nodes_dict)
    walker.register_listener(packer)
    walker.walk_graph(start_nodes)
    return PackedGraph(nodes_dict, tuple(adjacency_list))


def unpack_graph(packed_graph):
    """Return a list of GraphNodes from a PackedGraph

    Unpacks a PackedGraph, which must have the property: each parent node in
    the adjacency list appears after its children.
    """
    (nodes, adjacency_list) = packed_graph
    nodes_dict = dict(nodes)
    graph_nodes_dict = {}

    for (parent_sid, children_sids) in adjacency_list:
        if parent_sid in graph_nodes_dict:
            raise exception.InvalidInput(
                reason=_("PackedGraph adjacency list must be topologically "
                         "ordered"))
        children = []
        for child_sid in children_sids:
            if child_sid not in graph_nodes_dict:
                graph_nodes_dict[child_sid] = GraphNode(
                    nodes_dict[child_sid], ())
            children.append(graph_nodes_dict[child_sid])
            nodes_dict.pop(child_sid, None)
        graph_nodes_dict[parent_sid] = GraphNode(nodes_dict[parent_sid],
                                                 tuple(children))

    result_nodes = []
    for sid in nodes_dict:
        if sid not in graph_nodes_dict:
            graph_nodes_dict[sid] = GraphNode(nodes_dict[sid], ())
        result_nodes.append(graph_nodes_dict[sid])
    return result_nodes


def serialize_resource_graph(resource_graph):
    packed_resource_graph = pack_graph(resource_graph)
    return jsonutils.dumps(
        packed_resource_graph,
        default=lambda r: (r.type, r.id, r.name, r.extra_info))


def deserialize_resource_graph(serialized_resource_graph):
    deserialized_graph = jsonutils.loads(serialized_resource_graph)
    packed_resource_graph = PackedGraph(nodes=deserialized_graph[0],
                                        adjacency=deserialized_graph[1])
    for sid, node in packed_resource_graph.nodes.items():
        packed_resource_graph.nodes[sid] = Resource(type=node[0],
                                                    id=node[1],
                                                    name=node[2],
                                                    extra_info=node[3])
    resource_graph = unpack_graph(packed_resource_graph)
    return resource_graph



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from cinderclient.v3 import volumes
from collections import namedtuple
import mock

from karbor.common import constants
from karbor.context import RequestContext
from karbor.resource import Resource
from karbor.services.protection.protectable_plugins.volume \
    import VolumeProtectablePlugin

from karbor.tests import base
from oslo_config import cfg

project_info = namedtuple('project_info', field_names=['id', 'type', 'name'])
vol_info = namedtuple('vol_info', ['id', 'attachments', 'name', 'status',
                                   'availability_zone'])


class VolumeProtectablePluginTest(base.TestCase):
    def setUp(self):
        super(VolumeProtectablePluginTest, self).setUp()
        service_catalog = [
            {'type': 'volumev3',
             'endpoints': [{'publicURL': 'http://127.0.0.1:8776/v3/abcd'}],
             },
        ]
        self._context = RequestContext(user_id='demo',
                                       project_id='abcd',
                                       auth_token='efgh',
                                       service_catalog=service_catalog)

    def test_create_client_by_endpoint(self):
        cfg.CONF.set_default('cinder_endpoint',
                             'http://127.0.0.1:8776/v3',
                             'cinder_client')
        plugin = VolumeProtectablePlugin(self._context)
        self.assertEqual('volumev3',
                         plugin._client(self._context).client.service_type)
        self.assertEqual('http://127.0.0.1:8776/v3/abcd',
                         plugin._client(self._context).client.management_url)

    def test_create_client_by_catalog(self):
        plugin = VolumeProtectablePlugin(self._context)
        self.assertEqual('volumev3',
                         plugin._client(self._context).client.service_type)
        self.assertEqual('http://127.0.0.1:8776/v3/abcd',
                         plugin._client(self._context).client.management_url)

    def test_get_resource_type(self):
        plugin = VolumeProtectablePlugin(self._context)
        self.assertEqual("OS::Cinder::Volume", plugin.get_resource_type())

    def test_get_parent_resource_types(self):
        plugin = VolumeProtectablePlugin(self._context)
        self.assertItemsEqual(("OS::Nova::Server", "OS::Keystone::Project"),
                              plugin.get_parent_resource_types())

    @mock.patch.object(volumes.VolumeManager, 'list')
    def test_list_resources(self, mock_volume_list):
        plugin = VolumeProtectablePlugin(self._context)
        mock_volume_list.return_value = [
            vol_info('123', [], 'name123', 'available', 'az1'),
            vol_info('456', [], 'name456', 'available', 'az1'),
        ]
        self.assertEqual([Resource('OS::Cinder::Volume', '123', 'name123',
                                   {'availability_zone': 'az1'}),
                          Resource('OS::Cinder::Volume', '456', 'name456',
                                   {'availability_zone': 'az1'})],
                         plugin.list_resources(self._context))

    @mock.patch.object(volumes.VolumeManager, 'get')
    def test_show_resource(self, mock_volume_get):
        plugin = VolumeProtectablePlugin(self._context)

        vol_info = namedtuple('vol_info', ['id', 'name', 'status',
                              'availability_zone'])
        mock_volume_get.return_value = vol_info(id='123', name='name123',
                                                status='available',
                                                availability_zone='az1')
        self.assertEqual(Resource('OS::Cinder::Volume', '123', 'name123',
                                  {'availability_zone': 'az1'}),
                         plugin.show_resource(self._context, "123"))

    @mock.patch.object(volumes.VolumeManager, 'list')
    def test_get_server_dependent_resources(self, mock_volume_list):
        plugin = VolumeProtectablePlugin(self._context)

        attached = [{'server_id': 'abcdef', 'name': 'name'}]
        mock_volume_list.return_value = [
            vol_info('123', attached, 'name123', 'available', 'az1'),
            vol_info('456', [], 'name456', 'available', 'az1'),
        ]
        self.assertEqual([Resource('OS::Cinder::Volume', '123', 'name123',
                                   {'availability_zone': 'az1'})],
                         plugin.get_dependent_resources(
                             self._context,
                             Resource("OS::Nova::Server", 'abcdef', 'name',
                                      {'availability_zone': 'az1'})))

    @mock.patch.object(volumes.VolumeManager, 'list')
    def test_get_project_dependent_resources(self, mock_volume_list):
        project = project_info('abcd', constants.PROJECT_RESOURCE_TYPE,
                               'nameabcd')
        plugin = VolumeProtectablePlugin(self._context)

        volumes = [
            mock.Mock(name='Volume', id='123', availability_zone='az1'),
            mock.Mock(name='Volume', id='456', availability_zone='az1'),
        ]
        setattr(volumes[0], 'os-vol-tenant-attr:tenant_id', 'abcd')
        setattr(volumes[1], 'os-vol-tenant-attr:tenant_id', 'efgh')
        setattr(volumes[0], 'name', 'name123')
        setattr(volumes[1], 'name', 'name456')

        mock_volume_list.return_value = volumes
        self.assertEqual(
            [Resource('OS::Cinder::Volume', '123', 'name123',
                      {'availability_zone': 'az1'})],
            plugin.get_dependent_resources(self._context, project))



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Base Manager class.

Managers are responsible for a certain aspect of the system.  It is a logical
grouping of code relating to a portion of the system.  In general other
components should be using the manager to make changes to the components that
it is responsible for.

We have adopted a basic strategy of Smart managers and dumb data, which means
rather than attaching methods to data objects, components should call manager
methods that act on the data.

Methods on managers that can be executed locally should be called directly. If
a particular method must execute on a remote host, this should be done via rpc
to the service that wraps the manager

Managers should be responsible for most of the db access, and
non-implementation specific data.  Anything implementation specific that can't
be generalized should be done by the Driver.

Managers will often provide methods for initial setup of a host or periodic
tasks to a wrapping service.

This module provides Manager, a base class for managers.

"""


from oslo_config import cfg
import oslo_messaging as messaging
from oslo_service import periodic_task

from karbor.db import base
from karbor import version


CONF = cfg.CONF


class PeriodicTasks(periodic_task.PeriodicTasks):
    def __init__(self):
        super(PeriodicTasks, self).__init__(CONF)


class Manager(base.Base, PeriodicTasks):
    # Set RPC API version to 1.0 by default.
    RPC_API_VERSION = '1.0'

    target = messaging.Target(version=RPC_API_VERSION)

    def __init__(self, host=None, db_driver=None):
        if not host:
            host = CONF.host
        self.host = host
        self.additional_endpoints = []
        super(Manager, self).__init__(db_driver)

    def periodic_tasks(self, context, raise_on_error=False):
        """Tasks to be run at a periodic interval."""
        return self.run_periodic_tasks(context, raise_on_error=raise_on_error)

    def init_host(self, **kwargs):
        """Handle initialization if this is a standalone service.

        A hook point for services to execute tasks before the services are made
        available (i.e. showing up on RPC and starting to accept RPC calls) to
        other components.  Child classes should override this method.

        """
        pass

    def cleanup_host(self):
        """Hook to do cleanup work when the service shuts down.

        Child classes should override this method.
        """
        pass

    def init_host_with_rpc(self):
        """A hook for service to do jobs after RPC is ready.

        Like init_host(), this method is a hook where services get a chance
        to execute tasks that *need* RPC. Child classes should override
        this method.

        """
        pass

    def service_version(self):
        return version.version_string()

    def service_config(self):
        config = {}
        for key in CONF:
            config[key] = CONF.get(key, None)
        return config

    def is_working(self):
        """Method indicating if service is working correctly.

        This method is supposed to be overridden by subclasses and return if
        manager is working correctly.
        """
        return True



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from functools import partial
import mock

from karbor.common import constants
from karbor.resource import Resource
from karbor.services.protection.flows.workflow import TaskFlowEngine
from karbor.services.protection import graph
from karbor.services.protection import resource_flow
from karbor.services.protection import restore_heat
from karbor.tests import base
from karbor.tests.unit.protection import fakes
from oslo_config import cfg

CONF = cfg.CONF

(
    parent_type,
    child_type,
    grandchild_type,
) = fakes.FakeProtectionPlugin.SUPPORTED_RESOURCES

parent = Resource(id='A1', name='parent', type=parent_type)
child = Resource(id='B1', name='child', type=child_type)
grandchild = Resource(id='C1', name='grandchild', type=grandchild_type)


class ResourceFlowTest(base.TestCase):
    def setUp(self):
        super(ResourceFlowTest, self).setUp()

        self.resource_graph = {
            parent: [child],
            child: [grandchild],
            grandchild: [],
        }

        self.provider = fakes.FakeProvider()
        self.test_graph = graph.build_graph([parent],
                                            self.resource_graph.__getitem__)
        self.taskflow_engine = TaskFlowEngine()

    def _walk_operation(self, protection, operation_type,
                        checkpoint='checkpoint', parameters={}, context=None,
                        **kwargs):
        plugin_map = {
            parent_type: protection,
            child_type: protection,
            grandchild_type: protection,
        }
        flow = resource_flow.build_resource_flow(operation_type,
                                                 context,
                                                 self.taskflow_engine,
                                                 plugin_map,
                                                 self.test_graph,
                                                 parameters)

        store = {
            'checkpoint': checkpoint
        }
        store.update(kwargs)

        engine = self.taskflow_engine.get_engine(flow,
                                                 engine='parallel',
                                                 store=store)
        self.taskflow_engine.run_engine(engine)

    @mock.patch('karbor.tests.unit.protection.fakes.FakeProtectionPlugin')
    def test_resource_no_impl(self, mock_protection):
        for operation in constants.OPERATION_TYPES:
            kwargs = {}
            if operation == constants.OPERATION_RESTORE:
                kwargs['heat_template'] = restore_heat.HeatTemplate()
                kwargs['restore'] = None
            self._walk_operation(mock_protection, operation, **kwargs)

    @mock.patch('karbor.tests.unit.protection.fakes.FakeProtectionPlugin')
    def test_resource_flow_callbacks(self, mock_protection):
        for operation in constants.OPERATION_TYPES:
            mock_operation = fakes.MockOperation()
            get_operation_attr = 'get_{}_operation'.format(operation)
            getattr(
                mock_protection,
                get_operation_attr
            ).return_value = mock_operation

            kwargs = {}
            if operation == constants.OPERATION_RESTORE:
                kwargs['heat_template'] = restore_heat.HeatTemplate()
                kwargs['restore'] = None
            self._walk_operation(mock_protection, operation, **kwargs)

            self.assertEqual(mock_operation.on_prepare_begin.call_count,
                             len(self.resource_graph))
            self.assertEqual(mock_operation.on_prepare_finish.call_count,
                             len(self.resource_graph))
            self.assertEqual(mock_operation.on_main.call_count,
                             len(self.resource_graph))
            self.assertEqual(mock_operation.on_complete.call_count,
                             len(self.resource_graph))

    @mock.patch('karbor.tests.unit.protection.fakes.FakeProtectionPlugin')
    def test_resource_flow_parameters(self, mock_protection):
        resource_a1_id = "{}#{}".format(parent_type, 'A1')
        resource_b1_id = "{}#{}".format(child_type, 'B1')
        parameters = {
            resource_a1_id: {'option1': 'value1'},
            resource_b1_id: {'option2': 'value2', 'option3': 'value3'},
            parent_type: {'option4': 'value4'},
            child_type: {'option3': 'value5'}
        }

        def _compare_parameters(resource, func, expect_parameters):
            result = fake_operation.all_invokes[resource][func]
            for k, v in expect_parameters.items():
                self.assertEqual(v, result[k])

        for operation in constants.OPERATION_TYPES:
            fake_operation = fakes.FakeOperation()
            get_operation_attr = 'get_{}_operation'.format(operation)
            getattr(
                mock_protection,
                get_operation_attr
            ).return_value = fake_operation

            kwargs = {
                'checkpoint': 'A',
                'context': 'B',
            }

            args = kwargs.copy()
            if operation == constants.OPERATION_RESTORE:
                template = restore_heat.HeatTemplate()
                args['kwargs'] = {
                    'heat_template': template,
                    'restore': None
                }
                kwargs.update(args['kwargs'])

            self._walk_operation(mock_protection, operation,
                                 parameters=parameters, **kwargs)

            for resource in self.resource_graph:
                resource_params = parameters.get(resource.type, {})
                resource_id = '{}#{}'.format(resource.type, resource.id)
                resource_params.update(parameters.get(resource_id, {}))
                args['resource'] = resource
                args['parameters'] = resource_params
                _compare_parameters(resource, 'on_prepare_begin', args)
                _compare_parameters(resource, 'on_prepare_finish', args)
                _compare_parameters(resource, 'on_main', args)
                _compare_parameters(resource, 'on_complete', args)

    @mock.patch('karbor.tests.unit.protection.fakes.FakeProtectionPlugin')
    def test_resource_flow_order(self, mock_protection):
        def test_order(order_list, hook_type, resource, *args, **kwargs):
            order_list.append((hook_type, resource.id))

        operation = constants.OPERATION_PROTECT
        mock_operation = fakes.MockOperation()
        get_operation_attr = 'get_{}_operation'.format(operation)
        getattr(
            mock_protection,
            get_operation_attr
        ).return_value = mock_operation

        order_list = []
        mock_operation.on_prepare_begin = partial(test_order, order_list,
                                                  'pre_begin')
        mock_operation.on_prepare_finish = partial(test_order, order_list,
                                                   'pre_finish')
        mock_operation.on_main = partial(test_order, order_list, 'main')
        mock_operation.on_complete = partial(test_order, order_list,
                                             'complete')

        self._walk_operation(mock_protection, operation)

        self.assertLess(order_list.index(('pre_begin', parent.id)),
                        order_list.index(('pre_begin', child.id)))
        self.assertLess(order_list.index(('pre_begin', child.id)),
                        order_list.index(('pre_begin', grandchild.id)))

        self.assertGreater(order_list.index(('pre_finish', parent.id)),
                           order_list.index(('pre_finish', child.id)))
        self.assertGreater(order_list.index(('pre_finish', child.id)),
                           order_list.index(('pre_finish', grandchild.id)))

        self.assertGreater(order_list.index(('complete', parent.id)),
                           order_list.index(('complete', child.id)))
        self.assertGreater(order_list.index(('complete', child.id)),
                           order_list.index(('complete', grandchild.id)))

        for resource_id in (parent.id, child.id, grandchild.id):
            self.assertLess(order_list.index(('pre_begin', resource_id)),
                            order_list.index(('pre_finish', resource_id)))
            self.assertLess(order_list.index(('pre_finish', resource_id)),
                            order_list.index(('main', resource_id)))
            self.assertLess(order_list.index(('main', resource_id)),
                            order_list.index(('complete', resource_id)))



# Copyright 2011 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""RequestContext: context for requests that persist through all of karbor."""

import copy

from oslo_config import cfg
from oslo_context import context
from oslo_utils import timeutils
import six

from karbor.i18n import _
from karbor import policy

CONF = cfg.CONF


class RequestContext(context.RequestContext):
    """Security context and request information.

    Represents the user taking a given action within the system.

    """
    def __init__(self, user_id, project_id, is_admin=None, read_deleted="no",
                 roles=None, project_name=None, remote_address=None,
                 timestamp=None, request_id=None, auth_token=None,
                 overwrite=True, quota_class=None, service_catalog=None,
                 domain=None, user_domain=None, project_domain=None,
                 auth_token_info=None):
        """Initialize RequestContext.

        :param read_deleted: 'no' indicates deleted records are hidden, 'yes'
            indicates deleted records are visible, 'only' indicates that
            *only* deleted records are visible.

        :param overwrite: Set to False to ensure that the greenthread local
            copy of the index is not overwritten.
        """

        super(RequestContext, self).__init__(auth_token=auth_token,
                                             user=user_id,
                                             tenant=project_id,
                                             domain=domain,
                                             user_domain=user_domain,
                                             project_domain=project_domain,
                                             is_admin=is_admin,
                                             request_id=request_id)
        self.roles = roles or []
        self.project_name = project_name
        self.read_deleted = read_deleted
        self.remote_address = remote_address
        if not timestamp:
            timestamp = timeutils.utcnow()
        elif isinstance(timestamp, six.string_types):
            timestamp = timeutils.parse_isotime(timestamp)
        self.timestamp = timestamp
        self.quota_class = quota_class
        self._auth_token_info = auth_token_info

        if service_catalog:
            # Only include required parts of service_catalog
            self.service_catalog = [s for s in service_catalog
                                    if s.get('type') in
                                    ('identity', 'compute', 'object-store',
                                     'image', 'volume', 'volumev2', 'network',
                                     'volumev3', 'orchestration',
                                     'share', 'sharev2')]
        else:
            # if list is empty or none
            self.service_catalog = []

        # We need to have RequestContext attributes defined
        # when policy.check_is_admin invokes request logging
        # to make it loggable.
        if self.is_admin is None:
            self.is_admin = policy.check_is_admin(self.roles, self)
        elif self.is_admin and 'admin' not in self.roles:
            self.roles.append('admin')

    def _get_read_deleted(self):
        return self._read_deleted

    def _set_read_deleted(self, read_deleted):
        if read_deleted not in ('no', 'yes', 'only'):
            raise ValueError(_("read_deleted can only be one of 'no', "
                               "'yes' or 'only', not %r") % read_deleted)
        self._read_deleted = read_deleted

    def _del_read_deleted(self):
        del self._read_deleted

    read_deleted = property(_get_read_deleted, _set_read_deleted,
                            _del_read_deleted)

    def to_dict(self):
        result = super(RequestContext, self).to_dict()
        result['user_id'] = self.user_id
        result['project_id'] = self.project_id
        result['project_name'] = self.project_name
        result['domain'] = self.domain
        result['read_deleted'] = self.read_deleted
        result['roles'] = self.roles
        result['remote_address'] = self.remote_address
        result['timestamp'] = self.timestamp.isoformat()
        result['quota_class'] = self.quota_class
        result['service_catalog'] = self.service_catalog
        result['request_id'] = self.request_id
        result['auth_token_info'] = self._auth_token_info
        return result

    @classmethod
    def from_dict(cls, values):
        allowed_keys = [
            'user_id',
            'project_id',
            'project_name',
            'domain',
            'read_deleted',
            'remote_address',
            'timestamp',
            'quota_class',
            'service_catalog',
            'request_id',
            'is_admin',
            'roles',
            'auth_token',
            'user_domain',
            'project_domain',
            'auth_token_info'
        ]
        kwargs = {k: values[k] for k in values if k in allowed_keys}
        return cls(**kwargs)

    def to_policy_values(self):
        policy = super(RequestContext, self).to_policy_values()

        policy['is_admin'] = self.is_admin

        return policy

    def elevated(self, read_deleted=None, overwrite=False):
        """Return a version of this context with admin flag set."""
        context = self.deepcopy()
        context.is_admin = True

        if 'admin' not in context.roles:
            context.roles.append('admin')

        if read_deleted is not None:
            context.read_deleted = read_deleted

        return context

    def deepcopy(self):
        return copy.deepcopy(self)

    @property
    def project_id(self):
        return self.tenant

    @project_id.setter
    def project_id(self, value):
        self.tenant = value

    @property
    def user_id(self):
        return self.user

    @user_id.setter
    def user_id(self, value):
        self.user = value

    @property
    def auth_token_info(self):
        return self._auth_token_info


def get_admin_context(read_deleted="no"):
    return RequestContext(user_id=None,
                          project_id=None,
                          is_admin=True,
                          read_deleted=read_deleted,
                          overwrite=False)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from cinderclient import client as cinder_client
from glanceclient import client as glance_client
from karborclient import client as karbor_client
from manilaclient import client as manilaclient
from neutronclient.v2_0 import client as neutron_client
from novaclient import client as nova_client

import os_client_config

from oslotest import base


def _get_cloud_config(cloud='devstack'):
    return os_client_config.OpenStackConfig().get_one_cloud(cloud=cloud)


def _credentials(cloud='devstack'):
    """Retrieves credentials to run functional tests

    Credentials are either read via os-client-config from the environment
    or from a config file ('clouds.yaml'). Environment variables override
    those from the config file.

    devstack produces a clouds.yaml with two named clouds - one named
    'devstack' which has user privs and one named 'devstack-admin' which
    has admin privs. This function will default to getting the devstack-admin
    cloud as that is the current expected behavior.
    """
    return _get_cloud_config(cloud=cloud).get_auth_args()


def _get_endpoint(service_type):
    cloud_config = _get_cloud_config()
    keystone_session = cloud_config.get_session_client(service_type)
    keystone_auth = cloud_config.get_auth()
    region_name = cloud_config.get_region_name()
    return keystone_auth.get_endpoint(
        keystone_session,
        service_type=service_type,
        region_name=region_name,
    )


def _get_client_args(service_type, endpoint_type="publicURL"):
    cloud_config = _get_cloud_config()
    keystone_session = cloud_config.get_session_client(service_type)
    keystone_auth = cloud_config.get_auth()
    region_name = cloud_config.get_region_name()
    return {
        'session': keystone_session,
        'auth': keystone_auth,
        'service_type': service_type,
        'endpoint_type': endpoint_type,
        'region_name': region_name,
    }


def _get_karbor_client(api_version='1'):
    kwargs = _get_client_args('data-protect')
    client = karbor_client.Client(api_version, **kwargs)
    return client


def _get_cinder_client(api_version='3'):
    kwargs = _get_client_args('volumev3')
    client = cinder_client.Client(api_version, **kwargs)
    return client


def _get_manila_client(api_version='2'):
    kwargs = _get_client_args('sharev2')
    client = manilaclient.Client(api_version, **kwargs)
    return client


def _get_glance_client(api_version='2'):
    kwargs = _get_client_args('image')
    kwargs.pop('endpoint_type')
    client = glance_client.Client(api_version, **kwargs)
    return client


def _get_nova_client(api_version='2.26'):
    kwargs = _get_client_args('compute')
    client = nova_client.Client(api_version, **kwargs)
    return client


def _get_neutron_client(api_version='2'):
    kwargs = _get_client_args('network')
    client = neutron_client.Client(**kwargs)
    return client


class ObjectStore(object):
    """Stores objects for later closing.

    ObjectStore can be used to aggregate objects and close them.

    Example:

        with closing(ObjectStore()) as obj_store:
            obj = obj_store.store(SomeObject())

    or:

        obj_store = ObjectStore()
        obj_store.store(SomeObject())
        obj_store.close()
    """

    def __init__(self):
        super(ObjectStore, self).__init__()
        self._close_funcs = []

    def store(self, obj, close_func=None):
        self._close_funcs.append(close_func if close_func else obj.close)
        return obj

    def close(self):
        for close_func in reversed(self._close_funcs):
            close_func()


class KarborBaseTest(base.BaseTestCase):
    """Basic class for karbor fullstack testing.

    This class has common code shared for karbor fullstack testing
    including the various clients (karbor) and common
    setup/cleanup code.
    """

    def setUp(self):
        super(KarborBaseTest, self).setUp()
        self.cinder_client = _get_cinder_client()
        self.manila_client = _get_manila_client()
        self.glance_client = _get_glance_client()
        self.nova_client = _get_nova_client()
        self.neutron_client = _get_neutron_client()
        self.karbor_client = _get_karbor_client()
        self.keystone_endpoint = _get_endpoint('identity')
        self._testcase_store = ObjectStore()
        self.provider_id_noop = 'b766f37c-d011-4026-8228-28730d734a3f'
        self.provider_id_os = 'cf56bd3e-97a7-4078-b6d5-f36246333fd9'
        self.provider_id_fs_bank = '6659007d-6f66-4a0f-9cb4-17d6aded0bb9'
        self.provider_id_os_volume_snapshot = (
            '90d5bfea-a259-41e6-80c6-dcfcfcd9d827')

    def store(self, obj, close_func=None):
        return self._testcase_store.store(obj, close_func)

    def tearDown(self):
        self._testcase_store.close()
        super(KarborBaseTest, self).tearDown()

    def provider_list(self):
        return self.karbor_client.providers.list()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import six


@six.add_metaclass(abc.ABCMeta)
class ProtectablePlugin(object):
    """Base abstract class for protectable plugin.

    """

    def __init__(self, context=None):
        super(ProtectablePlugin, self).__init__()
        self._context = context

    def instance(self, context):
        return self.__class__(context)

    @abc.abstractmethod
    def get_resource_type(self):
        """Return the resource type that this plugin supports.

        Subclasses can implement as a classmethod
        """
        pass

    @abc.abstractmethod
    def get_parent_resource_types(self):
        """Return the possible parent resource types.

        Subclasses can implement as a classmethod
        """
        pass

    @abc.abstractmethod
    def list_resources(self, context, parameters=None):
        """List resource instances of type this plugin supported.

        :return: The list of resource instance.
        """
        pass

    @abc.abstractmethod
    def show_resource(self, context, resource_id, parameters=None):
        """Show resource detail information.

        """
        pass

    @abc.abstractmethod
    def get_dependent_resources(self, context, parent_resource):
        """List dependent resource instances.

        The listed resource instances are of type this plugin supported,
        and dependent by the given parent resource.

        :param parent_resource: the parent resource instance.
        :type parent_resource: one of parent resource types.
        :return: the list of dependent resource instances.
        """
        pass



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
from oslo_utils import uuidutils
from webob import exc

from karbor.api.v1 import plans as plan_api
from karbor.api.v1 import scheduled_operations as operation_api
from karbor.api.v1 import triggers as trigger_api
from karbor import context
from karbor import exception
from karbor.tests import base
from karbor.tests.unit.api import fakes
from karbor.tests.unit.api.v1 import test_triggers


class FakeRemoteOperationApi(object):
    def __init__(self):
        super(FakeRemoteOperationApi, self).__init__()
        self._create_operation_exception = None
        self._delete_operation_exception = None

    def create_scheduled_operation(self, context, operation):
        if self._create_operation_exception:
            raise self._create_operation_exception

    def delete_scheduled_operation(self, context, operation_id, trigger_id):
        if self._delete_operation_exception:
            raise self._delete_operation_exception


class ScheduledOperationApiTest(base.TestCase):

    def setUp(self):
        super(ScheduledOperationApiTest, self).setUp()

        self.remote_operation_api = FakeRemoteOperationApi()
        self.controller = operation_api.ScheduledOperationController()
        self.controller.operationengine_api = self.remote_operation_api

        self.ctxt = context.RequestContext('demo', 'fakeproject', True)
        self.req = fakes.HTTPRequest.blank('/v1/scheduled_operations')

        trigger = self._create_trigger()
        self._plan = self._create_plan(uuidutils.generate_uuid())
        self.default_create_operation_param = {
            "name": "123",
            "description": "123",
            "operation_type": "protect",
            "trigger_id": trigger['trigger_info']['id'],
            "operation_definition": {
                "plan_id": self._plan['id'],
                "provider_id": self._plan['provider_id']
            },
        }

    def test_create_operation_InvalidBody(self):
        self.assertRaises(exc.HTTPUnprocessableEntity,
                          self.controller.create,
                          self.req, {})

    def test_create_operation_InvalidName(self):
        body = self._get_create_operation_request_body()
        self.assertRaises(exc.HTTPBadRequest,
                          self.controller.create,
                          self.req, body)

    def test_create_operation_invalid_trigger(self):
        param = self.default_create_operation_param.copy()
        param['trigger_id'] = 123
        body = self._get_create_operation_request_body(param)
        self.assertRaises(exc.HTTPBadRequest,
                          self.controller.create,
                          self.req, body)

    def test_create_operation_receive_invalid_except(self):
        self.remote_operation_api._create_operation_exception =\
            exception.TriggerIsInvalid(trigger_id=None)

        param = self.default_create_operation_param.copy()
        body = self._get_create_operation_request_body(param)
        self.assertRaises(exc.HTTPBadRequest,
                          self.controller.create,
                          self.req, body)

        self.remote_operation_api._create_operation_exception = None

    def test_create_operation_receive_unknown_except(self):
        self.remote_operation_api._create_operation_exception =\
            exception.TriggerNotFound(id=None)

        param = self.default_create_operation_param.copy()
        body = self._get_create_operation_request_body(param)
        self.assertRaises(exc.HTTPInternalServerError,
                          self.controller.create,
                          self.req, body)

        self.remote_operation_api._create_operation_exception = None

    def test_create_operation(self):
        name = 'my protect'
        param = self.default_create_operation_param.copy()
        param['name'] = name
        body = self._get_create_operation_request_body(param)
        operation = self.controller.create(self.req, body)
        self.assertEqual(name, operation['scheduled_operation']['name'])

    def test_delete_operation_receive_NotFound_except(self):
        self.remote_operation_api._delete_operation_exception =\
            exception.ScheduledOperationStateNotFound(op_id=None)

        operation = self._create_one_operation()
        self.assertRaises(exc.HTTPInternalServerError,
                          self.controller.delete,
                          self.req,
                          operation['scheduled_operation']['id'])

        self.remote_operation_api._delete_operation_exception = None

    def test_delete_operation(self):
        operation = self._create_one_operation()
        self.controller.delete(self.req,
                               operation['scheduled_operation']['id'])
        self.assertRaises(exc.HTTPNotFound,
                          self.controller.show,
                          self.req,
                          operation['scheduled_operation']['id'])

    def test_show_operation_not_exist(self):
        self.assertRaises(exc.HTTPNotFound,
                          self.controller.show,
                          self.req,
                          '2a9ce1f3-cc1a-4516-9435-0ebb13caa398')

    def test_show_operation_invalid_id(self):
        self.assertRaises(exc.HTTPBadRequest,
                          self.controller.show,
                          self.req, 1)

    def test_show_operation(self):
        operation = self._create_one_operation()
        operation1 = self.controller.show(
            self.req, operation['scheduled_operation']['id'])
        self.assertEqual(operation['scheduled_operation']['id'],
                         operation1['scheduled_operation']['id'])

    def test_list_operation(self):
        operation = self._create_one_operation()
        operations = self.controller.index(self.req)
        for item in operations['operations']:
            if item['id'] == operation['scheduled_operation']['id']:
                self.assertTrue(1)

        self.assertFalse(0)

    def _create_one_operation(self):
        param = self.default_create_operation_param.copy()
        body = self._get_create_operation_request_body(param)
        return self.controller.create(self.req, body)

    def _get_create_operation_request_body(self, param={}):
        return {"scheduled_operation": param}

    def _create_trigger(self):
        create_trigger_param = {
            "trigger_info": {
                "name": "123",
                "type": "time",
                "properties": {
                    "format": "crontab",
                    "pattern": "* * * * *"
                },
            }
        }
        controller = trigger_api.TriggersController()
        controller.operationengine_api = test_triggers.FakeRemoteOperationApi()
        req = fakes.HTTPRequest.blank('/v1/triggers')
        return controller.create(req, create_trigger_param)

    @mock.patch(
        'karbor.services.protection.rpcapi.ProtectionAPI.show_provider')
    def _create_plan(self, provider_id, mock_provider):
        create_plan_param = {
            'plan': {
                'name': '123',
                'provider_id': provider_id,
                'resources': [
                    {'id': '39bb894794b741e982bd26144d2949f6',
                     'type': 'OS::Cinder::Volume', 'name': '123'}
                ],
                'parameters': {"OS::Cinder::Volume": {"backup_name": "test"}},
            }
        }
        controller = plan_api.PlansController()
        mock_provider.return_value = fakes.PROVIDER_OS
        req = fakes.HTTPRequest.blank('/v1/plans')
        plan = controller.create(req, create_plan_param)
        return plan['plan']



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
from oslo_serialization import jsonutils
from oslo_utils import timeutils

from karbor import objects
from karbor.tests.unit import objects as test_objects

NOW = timeutils.utcnow().replace(microsecond=0)

Trigger_ID = '0354ca9ddcd046b693340d78759fd274'

Fake_Trigger = {
    'created_at': NOW,
    'deleted_at': None,
    'updated_at': NOW,
    'deleted': False,
    'id': Trigger_ID,
    'name': 'daily',
    'project_id': '123',
    'type': 'time',
    'properties': '{}'
}


class TestTrigger(test_objects.BaseObjectsTestCase):
    Trigger_Class = objects.Trigger

    @mock.patch('karbor.db.trigger_get')
    def test_get_by_id(self, trigger_get):
        db_trigger = Fake_Trigger.copy()
        trigger_get.return_value = db_trigger

        trigger = self.Trigger_Class.get_by_id(self.context, Trigger_ID)
        db_trigger['properties'] = jsonutils.loads(db_trigger['properties'])
        self._compare(self, db_trigger, trigger)
        trigger_get.assert_called_once_with(self.context, Trigger_ID)

    @mock.patch('karbor.db.trigger_create')
    def test_create(self, trigger_create):
        db_trigger = Fake_Trigger.copy()
        trigger_create.return_value = db_trigger

        trigger = self.Trigger_Class(context=self.context)
        trigger.create()
        db_trigger['properties'] = jsonutils.loads(db_trigger['properties'])
        self._compare(self, db_trigger, trigger)
        trigger_create.assert_called_once_with(self.context, {})

    @mock.patch('karbor.db.trigger_update')
    def test_save(self, trigger_update):
        db_trigger = Fake_Trigger
        trigger = self.Trigger_Class._from_db_object(self.context,
                                                     self.Trigger_Class(),
                                                     db_trigger)
        trigger.name = 'weekly'
        trigger.save()

        trigger_update.assert_called_once_with(self.context,
                                               trigger.id,
                                               {'name': 'weekly'})

    @mock.patch('karbor.db.trigger_delete')
    def test_destroy(self, trigger_delete):
        db_trigger = Fake_Trigger
        trigger = self.Trigger_Class._from_db_object(self.context,
                                                     self.Trigger_Class(),
                                                     db_trigger)
        trigger.destroy()
        trigger_delete.assert_called_once_with(self.context, trigger.id)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Defines interface for DB access.

Functions in this module are imported into the karbor.db namespace. Call these
functions from karbor.db namespace, not the karbor.db.api namespace.

All functions in this module return objects that implement a dictionary-like
interface. Currently, many of these objects are sqlalchemy objects that
implement a dictionary interface. However, a future goal is to have all of
these objects be simple dictionaries.


**Related Flags**

:connection:  string specifying the sqlalchemy connection to use, like:
              `sqlite:///var/lib/karbor/karbor.sqlite`.

:enable_new_services:  when adding a new service to the database, is it in the
                       pool of available hardware (Default: True)

"""

from oslo_config import cfg
from oslo_db import concurrency as db_concurrency
from oslo_db import options as db_options


db_opts = [
    cfg.BoolOpt('enable_new_services',
                default=True,
                help='Services to be added to the available pool on create'),
]


CONF = cfg.CONF
CONF.register_opts(db_opts)
db_options.set_defaults(CONF)

_BACKEND_MAPPING = {'sqlalchemy': 'karbor.db.sqlalchemy.api'}


IMPL = db_concurrency.TpoolDbapiWrapper(CONF, _BACKEND_MAPPING)

# The maximum value a signed INT type may have
MAX_INT = 0x7FFFFFFF


###################

def dispose_engine():
    """Force the engine to establish new connections."""

    # FIXME(jdg): When using sqlite if we do the dispose
    # we seem to lose our DB here.  Adding this check
    # means we don't do the dispose, but we keep our sqlite DB
    # This likely isn't the best way to handle this

    if 'sqlite' not in IMPL.get_engine().name:
        return IMPL.dispose_engine()
    else:
        return


###################


def service_destroy(context, service_id):
    """Destroy the service or raise if it does not exist."""
    return IMPL.service_destroy(context, service_id)


def service_get(context, service_id):
    """Get a service or raise if it does not exist."""
    return IMPL.service_get(context, service_id)


def service_get_by_host_and_topic(context, host, topic):
    """Get a service by host it's on and topic it listens to."""
    return IMPL.service_get_by_host_and_topic(context, host, topic)


def service_get_all(context, disabled=None):
    """Get all services."""
    return IMPL.service_get_all(context, disabled)


def service_get_all_by_topic(context, topic, disabled=None):
    """Get all services for a given topic."""
    return IMPL.service_get_all_by_topic(context, topic, disabled=disabled)


def service_get_by_args(context, host, binary):
    """Get the state of an service by node name and binary."""
    return IMPL.service_get_by_args(context, host, binary)


def service_create(context, values):
    """Create a service from the values dictionary."""
    return IMPL.service_create(context, values)


def service_update(context, service_id, values):
    """Set the given properties on an service and update it.

    Raises NotFound if service does not exist.

    """
    return IMPL.service_update(context, service_id, values)


def get_by_id(context, model, id, *args, **kwargs):
    return IMPL.get_by_id(context, model, id, *args, **kwargs)


###################


def trigger_get(context, id):
    """Get a trigger by its id.

    :param context: The security context
    :param id: ID of the trigger

    :returns: Dictionary-like object containing properties of the trigger

    Raises TriggerNotFound if trigger with the given ID doesn't exist.
    """
    return IMPL.trigger_get(context, id)


def trigger_create(context, values):
    """Create a trigger from the values dictionary.

    :param context: The security context
    :param values: Dictionary containing trigger properties

    :returns: Dictionary-like object containing the properties of the created
              trigger
    """
    return IMPL.trigger_create(context, values)


def trigger_update(context, id, values):
    """Set the given properties on a trigger and update it.

    :param context: The security context
    :param id: ID of the trigger
    :param values: Dictionary containing trigger properties to be updated

    :returns: Dictionary-like object containing the properties of the updated
              trigger

    Raises TriggerNotFound if trigger with the given ID doesn't exist.
    """
    return IMPL.trigger_update(context, id, values)


def trigger_delete(context, id):
    """Delete a trigger from the database.

    :param context: The security context
    :param id: ID of the trigger

    Raises TriggerNotFound if trigger with the given ID doesn't exist.
    """
    return IMPL.trigger_delete(context, id)


def trigger_get_all_by_filters_sort(context, filters, limit=None,
                                    marker=None, sort_keys=None,
                                    sort_dirs=None):
    """Get all triggers that match all filters sorted by multiple keys.

    sort_keys and sort_dirs must be a list of strings.
    """
    return IMPL.trigger_get_all_by_filters_sort(
        context, filters, limit=limit, marker=marker,
        sort_keys=sort_keys, sort_dirs=sort_dirs)


###################


def scheduled_operation_get(context, id, columns_to_join=[]):
    """Get a scheduled operation by its id.

    :param context: The security context
    :param id: ID of the scheduled operation
    :param columns_to_join: columns which will be joined

    :returns: Dictionary-like object containing properties of the scheduled
     operation

    Raises ScheduledOperationNotFound if scheduled operation with
     the given ID doesn't exist.
    """
    return IMPL.scheduled_operation_get(context, id, columns_to_join)


def scheduled_operation_create(context, values):
    """Create a scheduled operation from the values dictionary.

    :param context: The security context
    :param values: Dictionary containing scheduled operation properties

    :returns: Dictionary-like object containing the properties of the created
              scheduled operation
    """
    return IMPL.scheduled_operation_create(context, values)


def scheduled_operation_update(context, id, values):
    """Set the given properties on a scheduled operation and update it.

    :param context: The security context
    :param id: ID of the scheduled operation
    :param values: Dictionary containing scheduled operation properties
                   to be updated

    :returns: Dictionary-like object containing the properties of the updated
              scheduled operation

    Raises ScheduledOperationNotFound if scheduled operation with
    the given ID doesn't exist.
    """
    return IMPL.scheduled_operation_update(context, id, values)


def scheduled_operation_delete(context, id):
    """Delete a scheduled operation from the database.

    :param context: The security context
    :param id: ID of the scheduled operation

    Raises ScheduledOperationNotFound if scheduled operation with
    the given ID doesn't exist.
    """
    return IMPL.scheduled_operation_delete(context, id)


def scheduled_operation_get_all_by_filters_sort(
        context, filters, limit=None,
        marker=None, sort_keys=None, sort_dirs=None):
    """Get all operations that match all filters sorted by multiple keys.

    sort_keys and sort_dirs must be a list of strings.
    """
    return IMPL.scheduled_operation_get_all_by_filters_sort(
        context, filters, limit=limit, marker=marker,
        sort_keys=sort_keys, sort_dirs=sort_dirs)


###################


def scheduled_operation_state_get(context, operation_id, columns_to_join=[]):
    """Get a scheduled operation state by its id.

    :param context: The security context
    :param operation_id: Operation_id of the scheduled operation state
    :columns_to_join: columns which will be joined

    :returns: Dictionary-like object containing properties of the scheduled
     operation state

    Raises ScheduledOperationStateNotFound if scheduled operation state with
     the given ID doesn't exist.
    """
    return IMPL.scheduled_operation_state_get(context, operation_id,
                                              columns_to_join)


def scheduled_operation_state_create(context, values):
    """Create a scheduled operation state from the values dictionary.

    :param context: The security context
    :param values: Dictionary containing scheduled operation state properties

    :returns: Dictionary-like object containing the properties of the created
              scheduled operation state
    """
    return IMPL.scheduled_operation_state_create(context, values)


def scheduled_operation_state_update(context, operation_id, values):
    """Set the given properties on a scheduled operation state and update it.

    :param context: The security context
    :param operation_id: Operation_id of the scheduled operation state
    :param values: Dictionary containing scheduled operation state properties
                   to be updated

    :returns: Dictionary-like object containing the properties of the updated
              scheduled operation state

    Raises ScheduledOperationStateNotFound if scheduled operation state with
    the given ID doesn't exist.
    """
    return IMPL.scheduled_operation_state_update(context, operation_id, values)


def scheduled_operation_state_delete(context, operation_id):
    """Delete a scheduled operation state from the database.

    :param context: The security context
    :param operation_id: Operation_id of the scheduled operation state

    Raises ScheduledOperationStateNotFound if scheduled operation state with
    the given ID doesn't exist.
    """
    return IMPL.scheduled_operation_state_delete(context, operation_id)


def scheduled_operation_state_get_all_by_filters_sort(
        context, filters, limit=None, marker=None,
        sort_keys=None, sort_dirs=None, columns_to_join=[]):
    """Get all operation states that match all filters sorted by multiple keys.

    sort_keys and sort_dirs must be a list of strings.
    """
    return IMPL.scheduled_operation_state_get_all_by_filters_sort(
        context, filters, limit=limit, marker=marker, sort_keys=sort_keys,
        sort_dirs=sort_dirs, columns_to_join=columns_to_join)


###################


def scheduled_operation_log_get(context, log_id):
    """Get a scheduled operation log by its id.

    :param context: The security context
    :param log_id: Log_id of the scheduled operation log

    :returns: Dictionary-like object containing properties of the scheduled
     operation log

    Raises ScheduledOperationLogNotFound if scheduled operation log with
     the given ID doesn't exist.
    """
    return IMPL.scheduled_operation_log_get(context, log_id)


def scheduled_operation_log_create(context, values):
    """Create a scheduled operation log from the values dictionary.

    :param context: The security context
    :param values: Dictionary containing scheduled operation log properties

    :returns: Dictionary-like object containing the properties of the created
              scheduled operation log
    """
    return IMPL.scheduled_operation_log_create(context, values)


def scheduled_operation_log_update(context, log_id, values):
    """Set the given properties on a scheduled operation log and update it.

    :param context: The security context
    :param log_id: Log_id of the scheduled operation log
    :param values: Dictionary containing scheduled operation log properties
                   to be updated

    :returns: Dictionary-like object containing the properties of the updated
              scheduled operation log

    Raises ScheduledOperationLogNotFound if scheduled operation log with
    the given ID doesn't exist.
    """
    return IMPL.scheduled_operation_log_update(context, log_id, values)


def scheduled_operation_log_delete(context, log_id):
    """Delete a scheduled operation log from the database.

    :param context: The security context
    :param log_id: Log_id of the scheduled operation log

    Raises ScheduledOperationLogNotFound if scheduled operation log with
    the given ID doesn't exist.
    """
    return IMPL.scheduled_operation_log_delete(context, log_id)


def scheduled_operation_log_delete_oldest(context, operation_id,
                                          retained_num, excepted_states=[]):
    """Delete the oldest scheduled operation logs from the database.

    :param context: The security context
    :param operation_id: ID of the scheduled operation
    :param retained_num: The number of retained logs
    :param excepted_states: If the state of log is in excepted_states,
                            it will not be deleted.
    """
    return IMPL.scheduled_operation_log_delete_oldest(context, operation_id,
                                                      retained_num,
                                                      excepted_states)


def scheduled_operation_log_get_all_by_filters_sort(
        context, filters, limit=None, marker=None,
        sort_keys=None, sort_dirs=None):
    """Get all operation logs that match all filters sorted by multiple keys.

    sort_keys and sort_dirs must be a list of strings.
    """
    return IMPL.scheduled_operation_log_get_all_by_filters_sort(
        context, filters, limit=limit, marker=marker, sort_keys=sort_keys,
        sort_dirs=sort_dirs)


###################


def plan_get(context, plan_id):
    """Get a plan or raise if it does not exist."""
    return IMPL.plan_get(context, plan_id)


def plan_create(context, values):
    """Create a plan from the values dictionary."""
    return IMPL.plan_create(context, values)


def plan_update(context, plan_id, values):
    """Set the given properties on a plan and update it.

    Raises NotFound if plan does not exist.

    """
    return IMPL.plan_update(context, plan_id, values)


def plan_resources_update(context, plan_id, resources):
    """Update resources if it exists, otherwise create it."""
    return IMPL.plan_resources_update(context, plan_id, resources)


def plan_destroy(context, plan_id):
    """Destroy the plan or raise if it does not exist."""
    return IMPL.plan_destroy(context, plan_id)


def plan_get_all(context, marker, limit, sort_keys=None, sort_dirs=None,
                 filters=None, offset=None):
    """Get all plans."""
    return IMPL.plan_get_all(context, marker, limit, sort_keys=sort_keys,
                             sort_dirs=sort_dirs, filters=filters,
                             offset=offset)


def plan_get_all_by_project(context, project_id, marker, limit,
                            sort_keys=None, sort_dirs=None, filters=None,
                            offset=None):
    """Get all plans belonging to a project."""
    return IMPL.plan_get_all_by_project(context, project_id, marker, limit,
                                        sort_keys=sort_keys,
                                        sort_dirs=sort_dirs,
                                        filters=filters,
                                        offset=offset)


def restore_get(context, restore_id):
    """Get a restore or raise if it does not exist."""
    return IMPL.restore_get(context, restore_id)


def restore_create(context, values):
    """Create a restore from the values dictionary."""
    return IMPL.restore_create(context, values)


def restore_update(context, restore_id, values):
    """Set the given properties on a restore and update it.

    Raises NotFound if plan does not exist.

    """
    return IMPL.restore_update(context, restore_id, values)


def restore_destroy(context, restore_id):
    """Destroy the restore or raise if it does not exist."""
    return IMPL.restore_destroy(context, restore_id)


def restore_get_all(context, marker, limit, sort_keys=None, sort_dirs=None,
                    filters=None, offset=None):
    """Get all restores."""
    return IMPL.restore_get_all(context, marker, limit, sort_keys=sort_keys,
                                sort_dirs=sort_dirs, filters=filters,
                                offset=offset)


def restore_get_all_by_project(context, project_id, marker, limit,
                               sort_keys=None, sort_dirs=None, filters=None,
                               offset=None):
    """Get all restores belonging to a project."""
    return IMPL.restore_get_all_by_project(context, project_id, marker, limit,
                                           sort_keys=sort_keys,
                                           sort_dirs=sort_dirs,
                                           filters=filters,
                                           offset=offset)


def operation_log_get(context, operation_log_id):
    """Get a operation log or raise if it does not exist."""
    return IMPL.operation_log_get(context, operation_log_id)


def operation_log_create(context, values):
    """Create a operation log from the values dictionary."""
    return IMPL.operation_log_create(context, values)


def operation_log_update(context, operation_log_id, values):
    """Set the given properties on a operation log and update it.

    Raises NotFound if plan does not exist.

    """
    return IMPL.operation_log_update(context, operation_log_id, values)


def operation_log_destroy(context, operation_log_id):
    """Destroy the operation log or raise if it does not exist."""
    return IMPL.operation_log_destroy(context, operation_log_id)


def operation_log_get_all(context, marker, limit, sort_keys=None,
                          sort_dirs=None,
                          filters=None, offset=None):
    """Get all operation logs."""
    return IMPL.operation_log_get_all(context, marker, limit,
                                      sort_keys=sort_keys,
                                      sort_dirs=sort_dirs,
                                      filters=filters,
                                      offset=offset)


def operation_log_get_all_by_project(context, project_id, marker, limit,
                                     sort_keys=None, sort_dirs=None,
                                     filters=None,
                                     offset=None):
    """Get all operation logs belonging to a project."""
    return IMPL.operation_log_get_all_by_project(context, project_id,
                                                 marker, limit,
                                                 sort_keys=sort_keys,
                                                 sort_dirs=sort_dirs,
                                                 filters=filters,
                                                 offset=offset)


###################


def checkpoint_record_get(context, checkpoint_record_id):
    """Get a checkpoint record or raise if it does not exist."""
    return IMPL.checkpoint_record_get(context, checkpoint_record_id)


def checkpoint_record_create(context, values):
    """Create a checkpoint record from the values dictionary."""
    return IMPL.checkpoint_record_create(context, values)


def checkpoint_record_update(context, checkpoint_record_id, values):
    """Set the given properties on a checkpoint record and update it.

    Raises NotFound if checkpoint record does not exist.

    """
    return IMPL.checkpoint_record_update(context, checkpoint_record_id, values)


def checkpoint_record_destroy(context, checkpoint_record_id):
    """Destroy the checkpoint record or raise if it does not exist."""
    return IMPL.checkpoint_record_destroy(context, checkpoint_record_id)


def checkpoint_record_get_all_by_filters_sort(
        context, filters, limit=None,
        marker=None, sort_keys=None, sort_dirs=None):
    """Get all checkpoint records that match all filters sorted

    by multiple keys. sort_keys and sort_dirs must be a list of strings.

    """
    return IMPL.checkpoint_record_get_all_by_filters_sort(
        context, filters, limit=limit, marker=marker,
        sort_keys=sort_keys, sort_dirs=sort_dirs)


def purge_deleted_rows(context, age_in_days):
    """Purge deleted rows older than given age from karbor tables

    Raises InvalidParameterValue if age_in_days is incorrect.
    :returns: number of deleted rows
    """
    return IMPL.purge_deleted_rows(context, age_in_days=age_in_days)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from keystoneauth1 import identity
from keystoneauth1 import session as keystone_session
from oslo_config import cfg
from swiftclient import client as swift

SERVICE = 'swift'
swift_client_opts = [
    cfg.StrOpt(SERVICE + '_endpoint',
               help='URL of the swift endpoint. Only used if '
                    'swift_auth_url is unset'),
    cfg.StrOpt(SERVICE + '_catalog_info',
               default='object-store:swift:publicURL',
               help='Info to match when looking for swift in the service '
                    'catalog. Format is: separated values of the form: '
                    '<service_type>:<service_name>:<endpoint_type> - '
                    'Only used if swift_endpoint and swift_auth_url '
                    'are unset'),
    cfg.StrOpt('swift_auth_url',
               help='The URL of the Keystone endpoint'),
    cfg.StrOpt('swift_tenant_name',
               help='Swift tenant/account name. '
                    'Required when connecting to an auth 2.0 system'),
    cfg.StrOpt('swift_user',
               help='Swift user name, if swift_auth_url is set.'),
    cfg.StrOpt('swift_key',
               secret=True,
               help='Swift key for authentication, if swift_auth_url '
                    ' is set.'),
    cfg.IntOpt('swift_retry_attempts',
               default=3,
               help='The number of retries to make for '
                    'Swift operations'),
    cfg.IntOpt('swift_retry_backoff',
               default=2,
               help='The backoff time in seconds '
                    'between Swift retries'),
    cfg.StrOpt('swift_ca_cert_file',
               help='Location of the CA certificate file '
                    'to use for swift client requests.'),
    cfg.BoolOpt('swift_auth_insecure',
                default=False,
                help='Bypass verification of server certificate when '
                     'making SSL connection to Swift.'),
]


def register_opts(conf):
    conf.register_opts(swift_client_opts, group=SERVICE + '_client')


def create(context, conf, **kwargs):
    register_opts(conf)

    client_config = conf.swift_client
    session = kwargs.get('session', None)

    if not session:
        auth = identity.Password(
            auth_url=client_config.swift_auth_url,
            username=client_config.swift_user,
            password=client_config.swift_key,
            project_name=client_config.swift_tenant_name,
        )
        session = keystone_session.Session(auth=auth)

    return swift.Connection(
        session=session,
        insecure=client_config.swift_auth_insecure,
        cacert=client_config.swift_ca_cert_file,
        retries=client_config.swift_retry_attempts,
        starting_backoff=client_config.swift_retry_backoff,
    )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from karbor.common import constants
from karbor.services.protection import protection_plugin


class NoopOperation(protection_plugin.Operation):
    def on_prepare_begin(self, *args, **kwargs):
        pass

    def on_prepare_finish(self, *args, **kwargs):
        pass

    def on_main(self, *args, **kwargs):
        pass

    def on_complete(self, *args, **kwargs):
        pass


class NoopProtectionPlugin(protection_plugin.ProtectionPlugin):
    def get_protect_operation(self, resource):
        return NoopOperation()

    def get_restore_operation(self, resource):
        return NoopOperation()

    def get_delete_operation(self, resource):
        return NoopOperation()

    @classmethod
    def get_supported_resources_types(cls):
        return constants.RESOURCE_TYPES

    @classmethod
    def get_options_schema(cls, resource_type):
        return {}

    @classmethod
    def get_saved_info_schema(cls, resource_type):
        return {}

    @classmethod
    def get_restore_schema(cls, resource_type):
        return {}

    @classmethod
    def get_saved_info(cls, metadata_store, resource):
        return None



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

from keystoneauth1 import loading
from oslo_config import cfg


CONF = cfg.CONF

CONF.import_opt('policy_file', 'karbor.policy', group='oslo_policy')
CONF.import_opt('provider_config_dir', 'karbor.services.protection.provider')


def set_defaults(conf):
    conf.set_default('connection', 'sqlite://', group='database')
    conf.set_default('sqlite_synchronous', False, group='database')
    conf.set_default('policy_file', 'karbor.tests.unit/policy.json',
                     group='oslo_policy')
    conf.set_default('policy_dirs', [], group='oslo_policy')
    conf.set_default('auth_strategy', 'noauth')
    conf.set_default('state_path', os.path.abspath(
        os.path.join(os.path.dirname(__file__), '..', '..', '..')))
    conf.set_default('provider_config_dir',
                     os.path.join(os.path.dirname(__file__), 'fake_providers'))
    loading.register_auth_conf_options(conf, 'trustee')
    opts = loading.get_auth_plugin_conf_options('password')
    conf.register_opts(opts, 'trustee')
    conf.set_default('auth_type', 'password', group='trustee')
    conf.set_default('auth_section', None, group='trustee')
    conf.set_default('auth_url', 'http://192.168.1.2/identity',
                     group='trustee')
    conf.set_default('username', 'karbor', group='trustee')
    conf.set_default('password', 'password', group='trustee')
    conf.set_default('user_domain_id', 'default', group='trustee')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

from karbor.services.protection import bank_plugin

fake_bank_opts = [
    cfg.HostAddressOpt('fake_host'),
]


class FakeBankPlugin(bank_plugin.BankPlugin):
    def __init__(self, config=None):
        super(FakeBankPlugin, self).__init__(config)
        config.register_opts(fake_bank_opts, 'fake_bank')
        self.fake_host = config['fake_bank']['fake_host']

    def update_object(self, key, value):
        return

    def get_object(self, key):
        return

    def list_objects(self, prefix=None, limit=None,
                     marker=None, sort_dir=None):
        return

    def delete_object(self, key):
        return

    def get_owner_id(self):
        return



#    Copyright 2015 SimpliVity Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from karbor import objects
from karbor.tests.unit import fake_restore
from karbor.tests.unit import objects as test_objects


class TestRestore(test_objects.BaseObjectsTestCase):
    @staticmethod
    def _compare(test, db, obj):
        db = {k: v for k, v in db.items()}
        test_objects.BaseObjectsTestCase._compare(test, db, obj)

    @mock.patch('karbor.objects.Restore.get_by_id')
    def test_get_by_id(self, restore_get):
        db_restore = fake_restore.fake_db_restore()
        restore_get.return_value = db_restore
        restore = objects.Restore.get_by_id(self.context, "1")
        restore_get.assert_called_once_with(self.context, "1")
        self._compare(self, db_restore, restore)

    @mock.patch('karbor.db.sqlalchemy.api.restore_create')
    def test_create(self, restore_create):
        db_restore = fake_restore.fake_db_restore()
        restore_create.return_value = db_restore
        restore = objects.Restore(context=self.context)
        restore.create()
        self.assertEqual(db_restore['id'], restore.id)

    @mock.patch('karbor.db.sqlalchemy.api.restore_update')
    def test_save(self, restore_update):
        db_restore = fake_restore.fake_db_restore()
        restore = objects.Restore._from_db_object(
            self.context, objects.Restore(), db_restore)
        restore.status = 'FAILED'
        restore.save()
        restore_update.assert_called_once_with(self.context, restore.id,
                                               {'status': 'FAILED'})

    @mock.patch('karbor.db.sqlalchemy.api.restore_destroy')
    def test_destroy(self, restore_destroy):
        db_restore = fake_restore.fake_db_restore()
        restore = objects.Restore._from_db_object(
            self.context, objects.Restore(), db_restore)
        restore.destroy()
        self.assertTrue(restore_destroy.called)
        admin_context = restore_destroy.call_args[0][0]
        self.assertTrue(admin_context.is_admin)

    def test_obj_field_status(self):
        restore = objects.Restore(context=self.context,
                                  status='FAILED')
        self.assertEqual('FAILED', restore.status)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from karbor import context
from karbor.services.operationengine import user_trust_manager
from karbor.tests import base


G_TOKEN_ID = 'abcdefg'
G_TRUST_ID = '1234556'


class FakeSession(object):
    def get_token(self):
        return G_TOKEN_ID


class FakeSKP(object):
    def create_trust_to_karbor(self, context):
        return G_TRUST_ID

    def delete_trust_to_karbor(self, trust_id):
        return

    def create_trust_session(self, trust_id):
        return FakeSession()


class UserTrustManagerTestCase(base.TestCase):
    """Test cases for UserTrustManager class."""

    def setUp(self):
        super(UserTrustManagerTestCase, self).setUp()

        self._user_id = '123'
        self._project_id = '456'
        self._ctx = context.RequestContext(user_id=self._user_id,
                                           project_id=self._project_id)
        self._manager = user_trust_manager.UserTrustManager()
        self._manager._skp = FakeSKP()

    def test_add_operation(self):
        manager = self._manager
        operation_id = 'abc'
        self.assertEqual(G_TRUST_ID, manager.add_operation(
            self._ctx, operation_id))

        info = manager._get_user_trust_info(self._user_id, self._project_id)
        self.assertIn(operation_id, info['operation_ids'])

        manager.add_operation(self._ctx, operation_id)
        self.assertEqual(1, len(info['operation_ids']))

    @mock.patch.object(FakeSKP, 'delete_trust_to_karbor')
    def test_delete_operation(self, del_trust):
        manager = self._manager
        op_ids = ['abc', '123']
        for op_id in op_ids:
            manager.add_operation(self._ctx, op_id)

        info = manager._get_user_trust_info(self._user_id, self._project_id)
        self.assertEqual(2, len(info['operation_ids']))

        manager.delete_operation(self._ctx, op_ids[0])
        self.assertEqual(1, len(info['operation_ids']))

        manager.delete_operation(self._ctx, op_ids[1])
        self.assertEqual(0, len(info['operation_ids']))
        del_trust.assert_called_once_with(G_TRUST_ID)

    def test_resume_operation(self):
        manager = self._manager
        operation_id = 'abc'
        manager.resume_operation(operation_id, self._user_id,
                                 self._project_id, G_TRUST_ID)

        info = manager._get_user_trust_info(self._user_id, self._project_id)
        self.assertIn(operation_id, info['operation_ids'])

        manager.resume_operation(operation_id, self._user_id,
                                 self._project_id, G_TRUST_ID)
        self.assertEqual(1, len(info['operation_ids']))

    def test_get_token(self):
        manager = self._manager
        manager.add_operation(self._ctx, 'abc')

        self.assertEqual(G_TOKEN_ID, manager.get_token(
            self._user_id, self._project_id))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import importutils

from karbor.common import constants
from karbor import exception
from karbor.services.protection.flows import delete as flow_delete
from karbor.services.protection.flows import protect as flow_protect
from karbor.services.protection.flows import restore as flow_restore

workflow_opts = [
    cfg.StrOpt(
        'workflow_engine',
        default="karbor.services.protection.flows.workflow.TaskFlowEngine",
        help='The workflow engine provides *flow* and *task* interface')
]

LOG = logging.getLogger(__name__)

CONF = cfg.CONF
CONF.register_opts(workflow_opts)


class Worker(object):
    def __init__(self, engine_path=None):
        super(Worker, self).__init__()
        try:
            self.workflow_engine = self._load_engine(engine_path)
        except Exception:
            LOG.error("load work flow engine failed")
            raise

    def _load_engine(self, engine_path):
        if not engine_path:
            engine_path = CONF.workflow_engine
        engine = importutils.import_object(engine_path)
        return engine

    def get_flow(self, context, operation_type, checkpoint, provider,
                 **kwargs):
        if operation_type == constants.OPERATION_PROTECT:
            plan = kwargs.get('plan', None)
            protectable_registry = kwargs.get('protectable_registry', None)
            flow = flow_protect.get_flow(
                context,
                protectable_registry,
                self.workflow_engine,
                plan,
                provider,
                checkpoint,
            )
        elif operation_type == constants.OPERATION_RESTORE:
            restore = kwargs.get('restore')
            restore_auth = kwargs.get('restore_auth')
            flow = flow_restore.get_flow(
                context,
                self.workflow_engine,
                checkpoint,
                provider,
                restore,
                restore_auth,
            )
        elif operation_type == constants.OPERATION_DELETE:
            flow = flow_delete.get_flow(
                context,
                self.workflow_engine,
                checkpoint,
                provider,
            )
        else:
            raise exception.InvalidParameterValue(
                err='unknown operation type %s' % operation_type
            )

        return flow

    def run_flow(self, flow_engine):
        self.workflow_engine.run_engine(flow_engine)

    def flow_outputs(self, flow_engine, target=None):
        return self.workflow_engine.output(flow_engine, target=target)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import re

from oslo_utils import importutils

from karbor.common import constants
from karbor.tests.fullstack import karbor_base
from karbor.tests.fullstack import karbor_objects as objects

DELETABLE_STATUS = {
    "volume_deletable_status": ["available", "error"],
    "server_deletable_status": ["available", "running"]
}


class RestoresTest(karbor_base.KarborBaseTest):
    """Test Restores operation """
    parameters = {}
    restore_auth = {
        "type": "password",
        "username": "admin",
        "password": "password",
    }

    def _store(self, resources_status):
        if not isinstance(resources_status, dict):
            return

        for resource, status in resources_status.items():
            resource_type, resource_id = resource.split("#")
            if resource_type is None:
                continue

            types = resource_type.split("::")
            if len(types) < 3:
                continue

            try:
                obj_class = importutils.import_class(
                    "karbor.tests.fullstack.karbor_objects.%s" % types[2])
            except Exception:
                continue

            deletable_str = "%s_deletable_status" % types[2].lower()
            deletable_list = eval(deletable_str, DELETABLE_STATUS)
            if callable(obj_class) and status in deletable_list:
                obj = obj_class()
                obj.id = resource_id
                obj.close()

    @staticmethod
    def get_restore_target(endpoint):
        regex = re.compile(
            r'http[s]?://\b(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}'
            r'(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\b', re.IGNORECASE
        )
        url = re.search(regex, endpoint).group()
        restore_target = url + r"/identity/v3"
        return restore_target

    def test_restore_create(self):
        volume = self.store(objects.Volume())
        volume.create(1)
        plan = self.store(objects.Plan())
        plan.create(self.provider_id_noop, [volume, ])
        checkpoint = self.store(objects.Checkpoint())
        checkpoint.create(self.provider_id_noop, plan.id)

        restore_target = self.get_restore_target(self.keystone_endpoint)
        restore = self.store(objects.Restore())
        restore.create(self.provider_id_noop, checkpoint.id,
                       restore_target, self.parameters, self.restore_auth)

        item = self.karbor_client.restores.get(restore.id)
        self.assertEqual(constants.RESTORE_STATUS_SUCCESS,
                         item.status)
        self._store(item.resources_status)

    def test_restore_create_without_target_and_auth(self):
        volume = self.store(objects.Volume())
        volume.create(1)
        plan = self.store(objects.Plan())
        plan.create(self.provider_id_noop, [volume, ])
        checkpoint = self.store(objects.Checkpoint())
        checkpoint.create(self.provider_id_noop, plan.id)

        restore = self.store(objects.Restore())
        restore.create(self.provider_id_noop, checkpoint.id,
                       None, self.parameters, None)

        item = self.karbor_client.restores.get(restore.id)
        self.assertEqual(constants.RESTORE_STATUS_SUCCESS,
                         item.status)
        self._store(item.resources_status)

    def test_restore_get(self):
        volume = self.store(objects.Volume())
        volume.create(1)
        plan = self.store(objects.Plan())
        plan.create(self.provider_id_noop, [volume, ])
        checkpoint = self.store(objects.Checkpoint())
        checkpoint.create(self.provider_id_noop, plan.id)

        restore_target = self.get_restore_target(self.keystone_endpoint)
        restore = self.store(objects.Restore())
        restore.create(self.provider_id_noop, checkpoint.id,
                       restore_target, self.parameters, self.restore_auth)

        item = self.karbor_client.restores.get(restore.id)
        self.assertEqual(restore.id, item.id)
        self.assertEqual(constants.RESTORE_STATUS_SUCCESS,
                         item.status)
        self._store(item.resources_status)

    def test_restore_list(self):
        volume = self.store(objects.Volume())
        volume.create(1)
        plan = self.store(objects.Plan())
        plan.create(self.provider_id_noop, [volume, ])
        checkpoint = self.store(objects.Checkpoint())
        checkpoint.create(self.provider_id_noop, plan.id)

        restores = self.karbor_client.restores.list()
        before_num = len(restores)

        restore_target = self.get_restore_target(self.keystone_endpoint)
        restore1 = self.store(objects.Restore())
        restore1.create(self.provider_id_noop, checkpoint.id,
                        restore_target, self.parameters, self.restore_auth)
        restore2 = self.store(objects.Restore())
        restore2.create(self.provider_id_noop, checkpoint.id,
                        restore_target, self.parameters, self.restore_auth)

        restores = self.karbor_client.restores.list()
        after_num = len(restores)
        self.assertLessEqual(2, after_num - before_num)

        item1 = self.karbor_client.restores.get(restore1.id)
        self._store(item1.resources_status)
        item2 = self.karbor_client.restores.get(restore2.id)
        self._store(item2.resources_status)

    def test_restore_resources(self):
        volume = self.store(objects.Volume())
        volume.create(1)
        plan = self.store(objects.Plan())
        plan.create(self.provider_id_os, [volume, ])
        checkpoint = self.store(objects.Checkpoint())
        checkpoint.create(self.provider_id_os, plan.id)

        restore_target = self.get_restore_target(self.keystone_endpoint)
        restore = self.store(objects.Restore())
        restore.create(self.provider_id_os, checkpoint.id,
                       restore_target, self.parameters, self.restore_auth)

        item = self.karbor_client.restores.get(restore.id)
        self.assertEqual(constants.RESTORE_STATUS_SUCCESS,
                         item.status)
        self.assertEqual(1, len(item.resources_status))
        self._store(item.resources_status)

    def test_restore_network_resources(self):
        network = self.store(objects.Network())
        network.create()
        plan = self.store(objects.Plan())
        plan.create(self.provider_id_os, [network, ])
        checkpoint = self.store(objects.Checkpoint())
        checkpoint.create(self.provider_id_os, plan.id)
        network.close()

        restore_target = self.get_restore_target(self.keystone_endpoint)
        restore = self.store(objects.Restore())
        restore.create(self.provider_id_os, checkpoint.id,
                       restore_target, self.parameters, self.restore_auth)

        item = self.karbor_client.restores.get(restore.id)
        self.assertEqual(constants.RESTORE_STATUS_SUCCESS,
                         item.status)
        self._store(item.resources_status)

    def test_restore_resources_with_fs_bank(self):
        volume = self.store(objects.Volume())
        volume.create(1)
        plan = self.store(objects.Plan())
        plan.create(self.provider_id_fs_bank, [volume, ])
        checkpoint = self.store(objects.Checkpoint())
        checkpoint.create(self.provider_id_fs_bank, plan.id)

        restore_target = self.get_restore_target(self.keystone_endpoint)
        restore = self.store(objects.Restore())
        restore.create(self.provider_id_fs_bank, checkpoint.id,
                       restore_target, self.parameters, self.restore_auth)

        item = self.karbor_client.restores.get(restore.id)
        self.assertEqual(constants.RESTORE_STATUS_SUCCESS,
                         item.status)
        self.assertEqual(1, len(item.resources_status))
        self._store(item.resources_status)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""The scheduled operations api."""

from oslo_log import log as logging
from oslo_utils import uuidutils
from webob import exc

from karbor.api import common
from karbor.api.openstack import wsgi
from karbor import exception
from karbor.i18n import _
from karbor import objects
from karbor import policy
from karbor.services.operationengine import api as operationengine_api
from karbor import utils

LOG = logging.getLogger(__name__)


def check_policy(context, action, target_obj=None):
    _action = 'scheduled_operation:%s' % action
    policy.enforce(context, _action, target_obj)


class ScheduledOperationViewBuilder(common.ViewBuilder):
    """Model a server API response as a python dictionary."""

    _collection_name = "scheduled_operations"

    def detail(self, request, operation):
        """Detailed view of a single scheduled operation."""

        operation_ref = {
            'scheduled_operation': {
                'id': operation.get('id'),
                'name': operation.get('name'),
                'description': operation.get('description'),
                'operation_type': operation.get('operation_type'),
                'trigger_id': operation.get('trigger_id'),
                'operation_definition': operation.get('operation_definition'),
                'enabled': operation.get('enabled'),
            }
        }
        return operation_ref

    def detail_list(self, request, operations):
        """Detailed view of a list of operations."""
        return self._list_view(self.detail, request, operations)

    def _list_view(self, func, request, operations):
        operations_list = [func(request, item)['scheduled_operation']
                           for item in operations]

        operations_links = self._get_collection_links(request,
                                                      operations,
                                                      self._collection_name,
                                                      )
        ret = {'operations': operations_list}
        if operations_links:
            ret['operations_links'] = operations_links

        return ret


class ScheduledOperationController(wsgi.Controller):
    """The Scheduled Operation API controller for the OpenStack API."""

    _view_builder_class = ScheduledOperationViewBuilder

    def __init__(self):
        self.operationengine_api = operationengine_api.API()
        super(ScheduledOperationController, self).__init__()

    def create(self, req, body):
        """Creates a new scheduled operation."""

        LOG.debug('Create scheduled operation start')

        if not self.is_valid_body(body, 'scheduled_operation'):
            raise exc.HTTPUnprocessableEntity()
        LOG.debug('Create a scheduled operation, request body: %s', body)

        context = req.environ['karbor.context']
        check_policy(context, 'create')
        operation_info = body['scheduled_operation']

        name = operation_info.get("name", None)
        operation_type = operation_info.get("operation_type", None)
        operation_definition = operation_info.get(
            "operation_definition", None)
        if not all([name, operation_type, operation_definition]):
            msg = _("Operation name or type or definition is not provided.")
            raise exc.HTTPBadRequest(explanation=msg)

        self.validate_name_and_description(operation_info)

        trigger_id = operation_info.get("trigger_id", None)
        trigger = self._get_trigger_by_id(context, trigger_id)
        if context.project_id != trigger.project_id:
            msg = _("Invalid trigger id provided.")
            raise exc.HTTPBadRequest(explanation=msg)

        operation_obj = {
            'name': operation_info.get('name', None),
            'description': operation_info.get('description', None),
            'operation_type': operation_type,
            'user_id': context.user_id,
            'project_id': context.project_id,
            'trigger_id': trigger_id,
            'operation_definition': operation_definition,
        }
        try:
            operation = objects.ScheduledOperation(context=context,
                                                   **operation_obj)
            operation.create()
        except Exception as ex:
            self._raise_unknown_exception(ex)

        try:
            self._create_scheduled_operation(context, operation)
        except Exception:
            try:
                operation.destroy()
            except Exception:
                pass

            raise

        return self._view_builder.detail(req, operation)

    def delete(self, req, id):
        """Delete a scheduled operation."""

        LOG.debug('Delete scheduled operation(%s) start', id)

        context = req.environ['karbor.context']
        operation = self._get_operation_by_id(context, id, ['trigger'])
        trigger = operation.trigger

        check_policy(context, 'delete', operation)

        try:
            self.operationengine_api.delete_scheduled_operation(
                context, id, trigger.id)

        except (exception.ScheduledOperationStateNotFound,
                exception.TriggerNotFound,
                Exception) as ex:
            self._raise_unknown_exception(ex)

        operation.destroy()

    def show(self, req, id):
        """Return data about the given operation."""

        LOG.debug('Get scheduled operation(%s) start', id)

        context = req.environ['karbor.context']
        operation = self._get_operation_by_id(context, id)
        check_policy(context, 'get', operation)

        return self._view_builder.detail(req, operation)

    def index(self, req):
        """Returns a list of operations, transformed through view builder."""

        context = req.environ['karbor.context']
        check_policy(context, 'list')

        params = req.params.copy()
        LOG.debug('List scheduled operation start, params=%s', params)
        marker, limit, offset = common.get_pagination_params(params)
        sort_keys, sort_dirs = common.get_sort_params(params)
        filters = params

        valid_filters = ["all_tenants", "name", "operation_type",
                         "trigger_id", "operation_definition"]
        utils.remove_invalid_filter_options(context, filters, valid_filters)
        utils.check_filters(filters)

        all_tenants = utils.get_bool_param("all_tenants", filters)
        if not (context.is_admin and all_tenants):
            filters["project_id"] = context.project_id

        try:
            operations = objects.ScheduledOperationList.get_by_filters(
                context, filters, limit, marker, sort_keys, sort_dirs)
        except Exception as ex:
            self._raise_unknown_exception(ex)

        return self._view_builder.detail_list(req, operations)

    def _get_operation_by_id(self, context, id, expect_attrs=[]):
        if not uuidutils.is_uuid_like(id):
            msg = _("Invalid operation id provided.")
            raise exc.HTTPBadRequest(explanation=msg)

        try:
            operation = objects.ScheduledOperation.get_by_id(
                context, id, expect_attrs)
        except exception.ScheduledOperationNotFound as error:
            raise exc.HTTPNotFound(explanation=error.msg)
        except Exception as ex:
            self._raise_unknown_exception(ex)

        return operation

    def _get_trigger_by_id(self, context, trigger_id):
        if not uuidutils.is_uuid_like(trigger_id):
            msg = _("Invalid trigger id provided.")
            raise exc.HTTPBadRequest(explanation=msg)
        try:
            trigger = objects.Trigger.get_by_id(context, trigger_id)
        except exception.NotFound as ex:
            raise exc.HTTPNotFound(explanation=ex.msg)
        except Exception as ex:
            self._raise_unknown_exception(ex)

        return trigger

    def _create_scheduled_operation(self, context, operation):
        try:
            self.operationengine_api.create_scheduled_operation(
                context, operation)

        except (exception.InvalidInput,
                exception.ScheduledOperationExist,
                exception.TriggerIsInvalid,
                exception.InvalidOperationDefinition) as ex:
            raise exc.HTTPBadRequest(explanation=ex.msg)

        except (exception.TriggerNotFound,
                exception.AuthorizationFailure,
                Exception) as ex:
            self._raise_unknown_exception(ex)

    def _raise_unknown_exception(self, exception_instance):
        LOG.exception('An unknown exception happened')

        value = exception_instance.msg if isinstance(
            exception_instance, exception.KarborException) else type(
            exception_instance)
        msg = (_('Unexpected API Error. Please report this at '
                 'http://bugs.launchpad.net/karbor/ and attach the '
                 'Karbor API log if possible.\n%s') % value)
        raise exc.HTTPInternalServerError(explanation=msg)


def create_resource():
    return wsgi.Resource(ScheduledOperationController())



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from karbor.common import constants
from karbor.resource import Resource
from karbor.services.protection import resource_flow
from oslo_log import log as logging
from taskflow import task

LOG = logging.getLogger(__name__)


class InitiateProtectTask(task.Task):
    def execute(self, checkpoint, *args, **kwargs):
        LOG.debug("Initiate protect checkpoint_id: %s", checkpoint.id)
        checkpoint.status = constants.CHECKPOINT_STATUS_PROTECTING
        checkpoint.commit()

    def revert(self, checkpoint, *args, **kwargs):
        LOG.debug("Failed to protect checkpoint_id: %s", checkpoint.id)
        checkpoint.status = constants.CHECKPOINT_STATUS_ERROR
        checkpoint.commit()


class CompleteProtectTask(task.Task):
    def execute(self, checkpoint):
        LOG.debug("Complete protect checkpoint_id: %s", checkpoint.id)
        checkpoint.status = constants.CHECKPOINT_STATUS_AVAILABLE
        checkpoint.commit()


def get_flow(context, protectable_registry, workflow_engine, plan, provider,
             checkpoint):
    resources = set(Resource(**item) for item in plan.get("resources"))
    resource_graph = protectable_registry.build_graph(context,
                                                      resources)
    checkpoint.resource_graph = resource_graph
    checkpoint.commit()
    flow_name = "Protect_" + plan.get('id')
    protection_flow = workflow_engine.build_flow(flow_name, 'linear')
    plugins = provider.load_plugins()
    resources_task_flow = resource_flow.build_resource_flow(
        operation_type=constants.OPERATION_PROTECT,
        context=context,
        workflow_engine=workflow_engine,
        resource_graph=resource_graph,
        plugins=plugins,
        parameters=plan.get('parameters'),
    )
    workflow_engine.add_tasks(
        protection_flow,
        InitiateProtectTask(),
        resources_task_flow,
        CompleteProtectTask(),
    )
    flow_engine = workflow_engine.get_engine(protection_flow, store={
        'checkpoint': checkpoint
    })
    return flow_engine



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import tempfile
import yaml

from karbor.services.protection.restore_heat import HeatResource
from karbor.services.protection.restore_heat import HeatTemplate
from karbor.tests import base


class HeatResourceTest(base.TestCase):
    def setUp(self):
        super(HeatResourceTest, self).setUp()
        fake_resource_id = "restore_123456"
        fake_resource_type = "OS::Cinder::Volume"
        self.heat_resource = HeatResource(fake_resource_id, fake_resource_type)

    def test_set_property(self):
        key = "volume_type"
        value = "lvmdriver-1"
        self.heat_resource.set_property(key, value)
        self.assertEqual({"volume_type": "lvmdriver-1"},
                         self.heat_resource._properties)

    def test_to_dict(self):
        properties = {
            "volume_type": "lvmdriver-1",
            "size": 1
        }
        for key, value in properties.items():
            self.heat_resource.set_property(key, value)
        resource_dict = self.heat_resource.to_dict()

        target_dict = {
            "restore_123456": {
                "type": "OS::Cinder::Volume",
                "properties": {
                    "volume_type": "lvmdriver-1",
                    "size": 1
                }
            }
        }
        self.assertEqual(target_dict, resource_dict)

    def tearDown(self):
        super(HeatResourceTest, self).tearDown()


class HeatTemplateTest(base.TestCase):
    def setUp(self):
        super(HeatTemplateTest, self).setUp()
        self.heat_template = HeatTemplate()

    def test_put_resource(self):
        fake_original_id = "123456"
        fake_resource_id = "restore_123456"
        fake_resource_type = "OS::Cinder::Volume"
        heat_resource = HeatResource(fake_resource_id, fake_resource_type)
        properties = {
            "volume_type": "lvmdriver-1",
            "size": 1
        }
        for key, value in properties.items():
            heat_resource.set_property(key, value)
        self.heat_template.put_resource(fake_original_id, heat_resource)
        self.assertEqual(1, len(self.heat_template._resources))
        self.assertEqual(
            fake_resource_id,
            self.heat_template._original_id_resource_map[fake_original_id]
        )

    def test_put_parameter(self):
        fake_original_id = "123456"
        fake_parameter = "restored_123456"
        self.heat_template.put_parameter(fake_original_id, fake_parameter)
        self.assertEqual(1, len(self.heat_template._original_id_parameter_map))
        self.assertEqual(
            fake_parameter,
            self.heat_template._original_id_parameter_map[fake_original_id]
        )

    def test_get_resource_reference(self):
        fake_original_id = "123456"
        fake_resource_id = "restore_123456"
        fake_resource_type = "OS::Cinder::Volume"
        heat_resource = HeatResource(fake_resource_id, fake_resource_type)
        properties = {
            "volume_type": "lvmdriver-1",
            "size": 1
        }
        for key, value in properties.items():
            heat_resource.set_property(key, value)
        self.heat_template.put_resource(fake_original_id, heat_resource)

        reference = self.heat_template.get_resource_reference(fake_original_id)
        self.assertEqual({"get_resource": "restore_123456"}, reference)

        fake_original_id = '23456'
        fake_parameter = 'restored_23456'
        self.heat_template.put_parameter(fake_original_id, fake_parameter)
        reference = self.heat_template.get_resource_reference(fake_original_id)
        self.assertEqual(fake_parameter, reference)

    def test_to_dict(self):
        fake_original_id = "123456"
        fake_resource_id = "restore_123456"
        fake_resource_type = "OS::Cinder::Volume"
        heat_resource = HeatResource(fake_resource_id, fake_resource_type)
        properties = {
            "volume_type": "lvmdriver-1",
            "size": 1
        }
        for key, value in properties.items():
            heat_resource.set_property(key, value)
        self.heat_template.put_resource(fake_original_id, heat_resource)

        fake_original_id_2 = '23456'
        fake_parameter = 'restored_23456'
        self.heat_template.put_parameter(fake_original_id_2, fake_parameter)

        template_dict = self.heat_template.to_dict()

        target_dict = {
            "heat_template_version": str(datetime.date(2015, 10, 15)),
            "description": "karbor restore template",
            "resources": {
                "restore_123456": {
                    "type": "OS::Cinder::Volume",
                    "properties": {
                        "volume_type": "lvmdriver-1",
                        "size": 1
                    }
                }
            }
        }
        self.assertEqual(target_dict, template_dict)

    def test_dump_to_yaml_file(self):
        fake_original_id = "123456"
        temp_dir = tempfile.mkdtemp()
        temp_file = temp_dir + "/template.yaml"

        fake_resource_id = "restore_123456"
        fake_resource_type = "OS::Cinder::Volume"
        heat_resource = HeatResource(fake_resource_id, fake_resource_type)
        properties = {
            "volume_type": "lvmdriver-1",
            "size": 1
        }
        for key, value in properties.items():
            heat_resource.set_property(key, value)
        self.heat_template.put_resource(fake_original_id, heat_resource)

        self.heat_template.dump_to_yaml_file(temp_file)

        with open(temp_file, "r") as f:
            template_dict = yaml.load(f)
        target_dict = {
            "heat_template_version": str(datetime.date(2015, 10, 15)),
            "description": "karbor restore template",
            "resources": {
                "restore_123456": {
                    "type": "OS::Cinder::Volume",
                    "properties": {
                        "volume_type": "lvmdriver-1",
                        "size": 1
                    }
                }
            }
        }
        self.assertEqual(target_dict, template_dict)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from datetime import datetime
from datetime import timedelta

from karbor import exception
from karbor.services.operationengine.engine.triggers.timetrigger.timeformats \
    import crontab_time
from karbor.tests import base


class CrontabTimeTestCase(base.TestCase):

    def setUp(self):
        super(CrontabTimeTestCase, self).setUp()

        self._time_format = crontab_time.Crontab

    def test_none_pattern(self):
        self.assertRaisesRegex(exception.InvalidInput,
                               "The trigger pattern is None",
                               self._time_format.check_time_format,
                               "")

    def test_invalid_pattern(self):
        self.assertRaisesRegex(exception.InvalidInput,
                               "The trigger pattern.* is invalid",
                               self._time_format.check_time_format,
                               "*")

    def test_compute_next_time(self):
        now = datetime(2016, 1, 20, 15, 11, 0, 0)
        obj = self._time_format(now, "* * * * *")
        time1 = obj.compute_next_time(now)
        time2 = now + timedelta(minutes=1)
        self.assertEqual(time2, time1)

    def test_get_interval(self):
        obj = self._time_format(datetime.now(), "* * * * *")
        self.assertEqual(60, obj.get_min_interval())



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from datetime import datetime
import mock

from karbor.common import constants
from karbor import context
from karbor import exception
from karbor import objects
from karbor.services.operationengine.operations import base as base_operation
from karbor.services.operationengine.operations import protect_operation
from karbor.tests import base


class FakeUserTrustManager(object):
    def add_operation(self, context, operation_id):
        return "123"

    def delete_operation(self, context, operation_id):
        pass

    def resume_operation(self, operation_id, user_id, project_id, trust_id):
        pass


class FakeCheckPoint(object):
    def create(self, provider_id, plan_id):
        return


class FakeKarborClient(object):
    def __init__(self):
        super(FakeKarborClient, self).__init__()
        self._check_point = FakeCheckPoint()

    @property
    def checkpoints(self):
        return self._check_point


class ProtectOperationTestCase(base.TestCase):
    """Test cases for ProtectOperation class."""

    def setUp(self):
        super(ProtectOperationTestCase, self).setUp()
        self._user_trust_manager = FakeUserTrustManager()
        self._operation = protect_operation.ProtectOperation(
            self._user_trust_manager
        )
        self._operation_db = self._create_operation()
        self._fake_karbor_client = FakeKarborClient()

    def test_check_operation_definition(self):
        self.assertRaises(exception.InvalidOperationDefinition,
                          self._operation.check_operation_definition,
                          {})

    @mock.patch.object(base_operation.Operation, '_create_karbor_client')
    def test_execute(self, client):
        client.return_value = self._fake_karbor_client
        now = datetime.utcnow()
        param = {
            'operation_id': self._operation_db.id,
            'triggered_time': now,
            'expect_start_time': now,
            'window_time': 30,
            'run_type': constants.OPERATION_RUN_TYPE_EXECUTE,
            'user_id': self._operation_db.user_id,
            'project_id': self._operation_db.project_id
        }
        self._operation.run(self._operation_db.operation_definition,
                            param=param)

        logs = objects.ScheduledOperationLogList.get_by_filters(
            context.get_admin_context(),
            {'state': constants.OPERATION_EXE_STATE_SUCCESS,
             'operation_id': self._operation_db.id}, 1,
            None, ['created_at'], ['desc'])

        self.assertIsNotNone(logs)
        log = logs.objects[0]
        self.assertTrue(now, log.triggered_time)

    @mock.patch.object(base_operation.Operation, '_create_karbor_client')
    def test_resume(self, client):
        log = self._create_operation_log(self._operation_db.id)
        client.return_value = self._fake_karbor_client
        now = datetime.utcnow()
        param = {
            'operation_id': self._operation_db.id,
            'triggered_time': now,
            'expect_start_time': now,
            'window_time': 30,
            'run_type': constants.OPERATION_RUN_TYPE_RESUME,
            'user_id': self._operation_db.user_id,
            'project_id': self._operation_db.project_id
        }
        self._operation.run(self._operation_db.operation_definition,
                            param=param)

        logs = objects.ScheduledOperationLogList.get_by_filters(
            context.get_admin_context(),
            {'state': constants.OPERATION_EXE_STATE_SUCCESS,
             'operation_id': self._operation_db.id}, 1,
            None, ['created_at'], ['desc'])

        self.assertIsNotNone(logs)
        log1 = logs.objects[0]
        self.assertTrue(log.id, log1.id)

    def _create_operation(self):
        operation_info = {
            'name': 'protect vm',
            'description': 'protect vm resource',
            'operation_type': 'protect',
            'user_id': '123',
            'project_id': '123',
            'trigger_id': '123',
            'operation_definition': {
                'provider_id': '123',
                'plan_id': '123'
            }
        }
        operation = objects.ScheduledOperation(context.get_admin_context(),
                                               **operation_info)
        operation.create()
        return operation

    def _create_operation_log(self, operation_id):
        log_info = {
            'operation_id': operation_id,
            'state': constants.OPERATION_EXE_STATE_IN_PROGRESS,
        }
        log = objects.ScheduledOperationLog(context.get_admin_context(),
                                            **log_info)
        log.create()
        return log



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests for Models Database."""

from datetime import datetime
from datetime import timedelta
from oslo_config import cfg
from oslo_utils import uuidutils
import six

from karbor import context
from karbor import db
from karbor import exception
from karbor.tests import base


CONF = cfg.CONF


class ModelBaseTestCase(base.TestCase):
    """Base Test cases which supplies assert Objects equal or not."""

    def _dict_from_object(self, obj, ignored_keys):
        if ignored_keys is None:
            ignored_keys = []
        if isinstance(obj, dict):
            items = obj.items()
        else:
            items = obj.iteritems()
        return {k: v for k, v in items
                if k not in ignored_keys}

    def _assertEqualObjects(self, obj1, obj2, ignored_keys=None):
        obj1 = self._dict_from_object(obj1, ignored_keys)
        obj2 = self._dict_from_object(obj2, ignored_keys)

        self.assertEqual(
            len(obj1), len(obj2),
            "Keys mismatch: %s" % six.text_type(
                set(obj1.keys()) ^ set(obj2.keys())))
        for key, value in obj1.items():
            self.assertEqual(value, obj2[key])


class ServicesDbTestCase(base.TestCase):
    """Test cases for Services database table."""

    def setUp(self):
        super(ServicesDbTestCase, self).setUp()
        self.ctxt = context.RequestContext(user_id='user_id',
                                           project_id='project_id',
                                           is_admin=True)

    def test_services_create(self):
        service_ref = db.service_create(self.ctxt,
                                        {'host': 'hosttest',
                                         'binary': 'binarytest',
                                         'topic': 'topictest',
                                         'report_count': 0})
        self.assertEqual('hosttest', service_ref['host'])

    def test_services_get(self):
        service_ref = db.service_create(self.ctxt,
                                        {'host': 'hosttest1',
                                         'binary': 'binarytest1',
                                         'topic': 'topictest1',
                                         'report_count': 0})

        service_get_ref = db.service_get(self.ctxt, service_ref['id'])
        self.assertEqual('hosttest1', service_ref['host'])
        self.assertEqual('hosttest1', service_get_ref['host'])

    def test_service_destroy(self):
        service_ref = db.service_create(self.ctxt,
                                        {'host': 'hosttest2',
                                         'binary': 'binarytest2',
                                         'topic': 'topictest2',
                                         'report_count': 0})
        service_id = service_ref['id']
        db.service_destroy(self.ctxt, service_id)
        self.assertRaises(exception.ServiceNotFound, db.service_get,
                          self.ctxt, service_id)

    def test_service_update(self):
        service_ref = db.service_create(self.ctxt,
                                        {'host': 'hosttest3',
                                         'binary': 'binarytest3',
                                         'topic': 'topictest3',
                                         'report_count': 0})
        service_id = service_ref['id']
        service_update_ref = db.service_update(self.ctxt, service_id,
                                               {'host': 'hosttest4',
                                                'binary': 'binarytest4',
                                                'topic': 'topictest4',
                                                'report_count': 0})
        self.assertEqual('hosttest3', service_ref['host'])
        self.assertEqual('hosttest4', service_update_ref['host'])

    def test_service_get_by_host_and_topic(self):
        service_ref = db.service_create(self.ctxt,
                                        {'host': 'hosttest5',
                                         'binary': 'binarytest5',
                                         'topic': 'topictest5',
                                         'report_count': 0})

        service_get_ref = db.service_get_by_host_and_topic(self.ctxt,
                                                           'hosttest5',
                                                           'topictest5')
        self.assertEqual('hosttest5', service_ref['host'])
        self.assertEqual('hosttest5', service_get_ref['host'])


class TriggerTestCase(base.TestCase):
    """Test cases for triggers table."""

    def setUp(self):
        super(TriggerTestCase, self).setUp()
        self.ctxt = context.RequestContext(user_id='user_id',
                                           project_id='project_id')

    def _create_trigger(self):
        values = {
            'id': "0354ca9ddcd046b693340d78759fd274",
            'name': 'first trigger',
            'project_id': self.ctxt.tenant,
            'type': 'time',
            'properties': '{}',
        }
        return db.trigger_create(self.ctxt, values)

    def test_trigger_create(self):
        trigger_ref = self._create_trigger()
        self.assertEqual('time', trigger_ref['type'])

    def test_trigger_delete(self):
        trigger_ref = self._create_trigger()
        db.trigger_delete(self.ctxt, trigger_ref['id'])

        self.assertRaises(exception.TriggerNotFound,
                          db.trigger_delete,
                          self.ctxt, trigger_ref['id'])

        self.assertRaises(exception.TriggerNotFound,
                          db.trigger_get,
                          self.ctxt, trigger_ref['id'])

        self.assertRaises(exception.TriggerNotFound,
                          db.trigger_delete, self.ctxt, '100')

    def test_trigger_update(self):
        trigger_ref = self._create_trigger()
        id = trigger_ref['id']
        trigger_ref = db.trigger_update(self.ctxt, id, {'type': 'event'})
        self.assertEqual('event', trigger_ref['type'])

        trigger_ref = db.trigger_get(self.ctxt, id)
        self.assertEqual('event', trigger_ref['type'])

        self.assertRaises(exception.TriggerNotFound,
                          db.trigger_update,
                          self.ctxt, '100', {"type": "event"})

    def test_trigger_get(self):
        trigger_ref = self._create_trigger()
        trigger_ref = db.trigger_get(self.ctxt, trigger_ref['id'])
        self.assertEqual('time', trigger_ref['type'])


class ScheduledOperationTestCase(base.TestCase):
    """Test cases for scheduled_operations table."""

    def setUp(self):
        super(ScheduledOperationTestCase, self).setUp()
        self.ctxt = context.RequestContext(user_id='user_id',
                                           project_id='project_id')

    def _create_scheduled_operation(self):
        values = {
            'id': '0354ca9ddcd046b693340d78759fd274',
            'name': 'protect vm',
            'description': 'protect vm resource',
            'operation_type': 'protect',
            'user_id': self.ctxt.user_id,
            'project_id': self.ctxt.tenant,
            'trigger_id': '0354ca9ddcd046b693340d78759fd275',
            'operation_definition': '{}'
        }
        return db.scheduled_operation_create(self.ctxt, values)

    def test_scheduled_operation_create(self):
        operation_ref = self._create_scheduled_operation()
        self.assertEqual('protect', operation_ref['operation_type'])
        self.assertTrue(operation_ref['enabled'])

    def test_scheduled_operation_delete(self):
        operation_ref = self._create_scheduled_operation()
        db.scheduled_operation_delete(self.ctxt, operation_ref['id'])

        self.assertRaises(exception.ScheduledOperationNotFound,
                          db.scheduled_operation_delete,
                          self.ctxt, operation_ref['id'])

        self.assertRaises(exception.ScheduledOperationNotFound,
                          db.scheduled_operation_get,
                          self.ctxt, operation_ref['id'])

        self.assertRaises(exception.ScheduledOperationNotFound,
                          db.scheduled_operation_delete, self.ctxt, '100')

    def test_scheduled_operation_update(self):
        operation_ref = self._create_scheduled_operation()
        id = operation_ref['id']
        operation_ref = db.scheduled_operation_update(self.ctxt,
                                                      id,
                                                      {"name": "abc"})
        self.assertEqual('abc', operation_ref['name'])

        operation_ref = db.scheduled_operation_get(self.ctxt, id)
        self.assertEqual('abc', operation_ref['name'])

        self.assertRaises(exception.ScheduledOperationNotFound,
                          db.scheduled_operation_update,
                          self.ctxt, '100', {"name": "abc"})

    def test_scheduled_operation_get(self):
        operation_ref = self._create_scheduled_operation()
        operation_ref = db.scheduled_operation_get(self.ctxt,
                                                   operation_ref['id'])
        self.assertEqual('protect', operation_ref['operation_type'])

    def test_scheduled_operation_get_join_trigger(self):
        def _create_trigger():
            values = {
                'id': "0354ca9ddcd046b693340d78759fd275",
                'name': 'first trigger',
                'project_id': self.ctxt.tenant,
                'type': 'time',
                'properties': '{}',
            }
            return db.trigger_create(self.ctxt, values)

        trigger_ref = _create_trigger()
        operation_ref = self._create_scheduled_operation()
        operation_ref = db.scheduled_operation_get(
            self.ctxt,
            operation_ref['id'],
            ['trigger'])
        self.assertEqual('protect', operation_ref['operation_type'])
        self.assertEqual(trigger_ref['type'], operation_ref.trigger['type'])


class ScheduledOperationStateTestCase(base.TestCase):
    """Test cases for scheduled_operation_states table."""

    def setUp(self):
        super(ScheduledOperationStateTestCase, self).setUp()
        self.ctxt = context.RequestContext(user_id='user_id',
                                           project_id='project_id')

    def _create_scheduled_operation_state(self):
        values = {
            'operation_id': '0354ca9ddcd046b693340d78759fd274',
            'service_id': 1,
            'trust_id': '123',
            'state': 'init',
        }
        return db.scheduled_operation_state_create(self.ctxt, values)

    def test_scheduled_operation_state_create(self):
        state_ref = self._create_scheduled_operation_state()
        self.assertEqual('init', state_ref['state'])

    def test_scheduled_operation_state_delete(self):
        state_ref = self._create_scheduled_operation_state()
        db.scheduled_operation_state_delete(self.ctxt,
                                            state_ref['operation_id'])

        self.assertRaises(exception.ScheduledOperationStateNotFound,
                          db.scheduled_operation_state_delete,
                          self.ctxt, state_ref['operation_id'])

        self.assertRaises(exception.ScheduledOperationStateNotFound,
                          db.scheduled_operation_state_get,
                          self.ctxt, state_ref['operation_id'])

        self.assertRaises(exception.ScheduledOperationStateNotFound,
                          db.scheduled_operation_state_delete, self.ctxt, 100)

    def test_scheduled_operation_state_update(self):
        state_ref = self._create_scheduled_operation_state()
        operation_id = state_ref['operation_id']
        state_ref = db.scheduled_operation_state_update(self.ctxt,
                                                        operation_id,
                                                        {"state": "success"})
        self.assertEqual('success', state_ref['state'])

        state_ref = db.scheduled_operation_state_get(self.ctxt, operation_id)
        self.assertEqual('success', state_ref['state'])

        self.assertRaises(exception.ScheduledOperationStateNotFound,
                          db.scheduled_operation_state_update,
                          self.ctxt, '100', {"state": "success"})

    def test_scheduled_operation_state_get(self):
        state_ref = self._create_scheduled_operation_state()
        state_ref = db.scheduled_operation_state_get(self.ctxt,
                                                     state_ref['operation_id'])
        self.assertEqual('init', state_ref['state'])

    def test_scheduled_operation_state_get_join_operation(self):
        def _create_scheduled_operation():
            values = {
                'id': '0354ca9ddcd046b693340d78759fd274',
                'name': 'protect vm',
                'operation_type': 'protect',
                'user_id': self.ctxt.user_id,
                'project_id': self.ctxt.tenant,
                'trigger_id': '0354ca9ddcd046b693340d78759fd275',
                'operation_definition': '{}'
            }
            return db.scheduled_operation_create(self.ctxt, values)

        operation_ref = _create_scheduled_operation()
        self._create_scheduled_operation_state()
        state_ref = db.scheduled_operation_state_get(
            self.ctxt,
            operation_ref['id'],
            ['operation'])
        self.assertEqual(operation_ref['id'], state_ref.operation['id'])


class ScheduledOperationLogTestCase(base.TestCase):
    """Test cases for scheduled_operation_logs table."""

    def setUp(self):
        super(ScheduledOperationLogTestCase, self).setUp()
        self.ctxt = context.get_admin_context()
        self.operation_id = '0354ca9ddcd046b693340d78759fd274'

    def _create_scheduled_operation_log(self, state='in_progress',
                                        created_at=datetime.now()):
        values = {
            'operation_id': self.operation_id,
            'state': state,
            'created_at': created_at
        }
        return db.scheduled_operation_log_create(self.ctxt, values)

    def test_scheduled_operation_log_create(self):
        log_ref = self._create_scheduled_operation_log()
        self.assertEqual('in_progress', log_ref['state'])

    def test_scheduled_operation_log_delete(self):
        log_ref = self._create_scheduled_operation_log()
        db.scheduled_operation_log_delete(self.ctxt, log_ref['id'])

        self.assertRaises(exception.ScheduledOperationLogNotFound,
                          db.scheduled_operation_log_delete,
                          self.ctxt, log_ref['id'])

        self.assertRaises(exception.ScheduledOperationLogNotFound,
                          db.scheduled_operation_log_get,
                          self.ctxt, log_ref['id'])

        self.assertRaises(exception.ScheduledOperationLogNotFound,
                          db.scheduled_operation_log_delete,
                          self.ctxt, 100)

    def test_scheduled_operation_log_delete_oldest(self):
        log_ids = []
        states = ['success', 'in_progress', 'success', 'success']
        for i in range(4):
            t = datetime.now() + timedelta(hours=i)
            log = self._create_scheduled_operation_log(
                states[i], t)
            log_ids.append(log['id'])

        db.scheduled_operation_log_delete_oldest(
            self.ctxt, self.operation_id, 3)
        self.assertRaises(exception.ScheduledOperationLogNotFound,
                          db.scheduled_operation_log_get,
                          self.ctxt, log_ids[0])

        db.scheduled_operation_log_delete_oldest(
            self.ctxt, self.operation_id, 1, ['in_progress'])
        log_ref = db.scheduled_operation_log_get(self.ctxt, log_ids[1])
        self.assertEqual('in_progress', log_ref['state'])
        self.assertRaises(exception.ScheduledOperationLogNotFound,
                          db.scheduled_operation_log_get,
                          self.ctxt, log_ids[2])

    def test_scheduled_operation_log_update(self):
        log_ref = self._create_scheduled_operation_log()
        log_id = log_ref['id']
        log_ref = db.scheduled_operation_log_update(self.ctxt,
                                                    log_id,
                                                    {"state": "success"})
        self.assertEqual('success', log_ref['state'])

        log_ref = db.scheduled_operation_log_get(self.ctxt, log_id)
        self.assertEqual('success', log_ref['state'])

        self.assertRaises(exception.ScheduledOperationLogNotFound,
                          db.scheduled_operation_log_update,
                          self.ctxt, 100, {"state": "success"})

    def test_scheduled_operation_log_get(self):
        log_ref = self._create_scheduled_operation_log()
        log_ref = db.scheduled_operation_log_get(self.ctxt, log_ref['id'])
        self.assertEqual('in_progress', log_ref['state'])


class PlanDbTestCase(ModelBaseTestCase):
    """Unit tests for karbor.db.api.plan_*."""

    fake_plan = {
        'name': 'My 3 tier application',
        'description': 'My 3 tier application protection plan',
        'provider_id': 'efc6a88b-9096-4bb6-8634-cda182a6e12a',
        'status': 'suspended',
        'project_id': '39bb894794b741e982bd26144d2949f6',
        'resources': [],
        'parameters': '{OS::Nova::Server: {consistency: os}}'
    }

    fake_plan_with_resources = {
        'name': 'My 3 tier application',
        'description': 'My 3 tier application protection plan',
        'provider_id': 'efc6a88b-9096-4bb6-8634-cda182a6e12a',
        'status': 'suspended',
        'project_id': '39bb894794b741e982bd26144d2949f6',
        'resources': [{
            "id": "64e51e85-4f31-441f-9a5d-6e93e3196628",
            "type": "OS::Nova::Server",
            "name": "vm1"}],
        'parameters': '{OS::Nova::Server: {consistency: os}}'
    }

    def setUp(self):
        super(PlanDbTestCase, self).setUp()
        self.ctxt = context.get_admin_context()

    def test_plan_create(self):
        plan = db.plan_create(self.ctxt, self.fake_plan)
        self.assertTrue(uuidutils.is_uuid_like(plan['id']))
        self.assertEqual('suspended', plan.status)

    def test_plan_get(self):
        plan = db.plan_create(self.ctxt,
                              self.fake_plan)
        self._assertEqualObjects(plan, db.plan_get(self.ctxt,
                                                   plan['id']),
                                 ignored_keys='resources')

    def test_plan_destroy(self):
        plan = db.plan_create(self.ctxt, self.fake_plan)
        db.plan_destroy(self.ctxt, plan['id'])
        self.assertRaises(exception.PlanNotFound, db.plan_get,
                          self.ctxt, plan['id'])

    def test_plan_update(self):
        plan = db.plan_create(self.ctxt, self.fake_plan)
        db.plan_update(self.ctxt, plan['id'],
                       {'status': 'started'})
        plan = db.plan_get(self.ctxt, plan['id'])
        self.assertEqual('started', plan['status'])

    def test_plan_update_nonexistent(self):
        self.assertRaises(exception.PlanNotFound, db.plan_update,
                          self.ctxt, 42, {})

    def test_plan_resources_update(self):
        resources2 = [{
            "id": "61e51e85-4f31-441f-9a5d-6e93e3194444",
            "type": "OS::Cinder::Volume",
            "name": "vm2",
            "extra_info": "{'availability_zone': 'az1'}"}]

        plan = db.plan_create(self.ctxt, self.fake_plan)
        db_meta = db.plan_resources_update(self.ctxt, plan["id"], resources2)

        self.assertEqual("OS::Cinder::Volume", db_meta[0]["resource_type"])
        self.assertEqual("vm2", db_meta[0]["resource_name"])
        self.assertEqual("{'availability_zone': 'az1'}",
                         db_meta[0]["resource_extra_info"])


class RestoreDbTestCase(ModelBaseTestCase):
    """Unit tests for karbor.db.api.restore_*."""

    fake_restore = {
        "id": "36ea41b2-c358-48a7-9117-70cb7617410a",
        "project_id": "586cc6ce-e286-40bd-b2b5-dd32694d9944",
        "provider_id": "2220f8b1-975d-4621-a872-fa9afb43cb6c",
        "checkpoint_id": "09edcbdc-d1c2-49c1-a212-122627b20968",
        "restore_target": "192.168.1.2/identity/",
        "parameters": "{'username': 'admin'}",
        "status": "SUCCESS"
    }

    def setUp(self):
        super(RestoreDbTestCase, self).setUp()
        self.ctxt = context.get_admin_context()

    def test_restore_create(self):
        restore = db.restore_create(self.ctxt, self.fake_restore)
        self.assertTrue(uuidutils.is_uuid_like(restore['id']))
        self.assertEqual('SUCCESS', restore.status)

    def test_restore_get(self):
        restore = db.restore_create(self.ctxt,
                                    self.fake_restore)
        self._assertEqualObjects(restore, db.restore_get(self.ctxt,
                                                         restore['id']))

    def test_restore_destroy(self):
        restore = db.restore_create(self.ctxt, self.fake_restore)
        db.restore_destroy(self.ctxt, restore['id'])
        self.assertRaises(exception.RestoreNotFound, db.restore_get,
                          self.ctxt, restore['id'])

    def test_restore_update(self):
        restore = db.restore_create(self.ctxt, self.fake_restore)
        db.restore_update(self.ctxt, restore['id'],
                          {'status': 'INIT'})
        restore = db.restore_get(self.ctxt, restore['id'])
        self.assertEqual('INIT', restore['status'])

    def test_restore_update_nonexistent(self):
        self.assertRaises(exception.RestoreNotFound, db.restore_update,
                          self.ctxt, 42, {})


class OperationLogTestCase(ModelBaseTestCase):
    """Unit tests for karbor.db.api.operation_log_*."""

    fake_operation_log = {
        "id": "36ea41b2-c358-48a7-9117-70cb7617410a",
        "project_id": "586cc6ce-e286-40bd-b2b5-dd32694d9944",
        "scheduled_operation_id": "2220f8b1-975d-4621-a872-fa9afb43cb6c",
        "state": "failed",
        "error": "Could not access bank",
        "entries": "[entries:{'timestamp': '2015-08-27T09:50:51-05:00',"
                   "'message': 'Doing things'}]"
    }

    def setUp(self):
        super(OperationLogTestCase, self).setUp()
        self.ctxt = context.get_admin_context()

    def test_operation_log_create(self):
        operation_log = db.operation_log_create(self.ctxt,
                                                self.fake_operation_log)
        self.assertTrue(uuidutils.is_uuid_like(operation_log['id']))
        self.assertEqual('failed', operation_log.state)

    def test_operation_log_get(self):
        operation_log = db.operation_log_create(self.ctxt,
                                                self.fake_operation_log)
        self._assertEqualObjects(operation_log, db.operation_log_get(
            self.ctxt, operation_log['id']))

    def test_operation_log_destroy(self):
        operation_log = db.operation_log_create(self.ctxt,
                                                self.fake_operation_log)
        db.operation_log_destroy(self.ctxt, operation_log['id'])
        self.assertRaises(exception.OperationLogNotFound, db.operation_log_get,
                          self.ctxt, operation_log['id'])

    def test_operation_log_update(self):
        operation_log = db.operation_log_create(self.ctxt,
                                                self.fake_operation_log)
        db.operation_log_update(self.ctxt, operation_log['id'],
                                {'state': 'finished'})
        operation_log = db.operation_log_get(self.ctxt, operation_log['id'])
        self.assertEqual('finished', operation_log['state'])

    def test_operation_log_update_nonexistent(self):
        self.assertRaises(exception.OperationLogNotFound,
                          db.operation_log_update,
                          self.ctxt, 42, {})


class CheckpointRecordTestCase(ModelBaseTestCase):
    """Unit tests for karbor.db.api.checkpoint_record_*."""

    fake_checkpoint_record = {
        "id": "36ea41b2-c358-48a7-9117-70cb7617410a",
        "project_id": "586cc6ce-e286-40bd-b2b5-dd32694d9944",
        "checkpoint_id": "2220f8b1-975d-4621-a872-fa9afb43cb6c",
        "checkpoint_status": "available",
        "provider_id": "39bb894794b741e982bd26144d2949f6",
        "plan_id": "efc6a88b-9096-4bb6-8634-cda182a6e12b",
        "operation_id": "64e51e85-4f31-441f-9a5d-6e93e3196628",
        "create_by": "operation-engine",
        "extend_info": "[{"
                       "'id': '0354ca9d-dcd0-46b6-9334-0d78759fd275',"
                       "'type': 'OS::Nova::Server',"
                       "'name': 'vm1'"
                        "}]"
    }

    def setUp(self):
        super(CheckpointRecordTestCase, self).setUp()
        self.ctxt = context.get_admin_context()

    def test_checkpoint_record_create(self):
        checkpoint_record = db.checkpoint_record_create(
            self.ctxt,
            self.fake_checkpoint_record)
        self.assertTrue(uuidutils.is_uuid_like(checkpoint_record['id']))
        self.assertEqual('available', checkpoint_record.checkpoint_status)

    def test_checkpoint_record_get(self):
        checkpoint_record = db.checkpoint_record_create(
            self.ctxt,
            self.fake_checkpoint_record)
        self._assertEqualObjects(checkpoint_record, db.checkpoint_record_get(
            self.ctxt, checkpoint_record['id']))

    def test_checkpoint_record_destroy(self):
        checkpoint_record = db.checkpoint_record_create(
            self.ctxt,
            self.fake_checkpoint_record)
        db.checkpoint_record_destroy(self.ctxt, checkpoint_record['id'])
        self.assertRaises(exception.CheckpointRecordNotFound,
                          db.checkpoint_record_get,
                          self.ctxt, checkpoint_record['id'])

    def test_checkpoint_record_update(self):
        checkpoint_record = db.checkpoint_record_create(
            self.ctxt,
            self.fake_checkpoint_record)
        db.checkpoint_record_update(self.ctxt,
                                    checkpoint_record['id'],
                                    {'checkpoint_status': 'error'})
        checkpoint_record = db.checkpoint_record_get(
            self.ctxt,
            checkpoint_record['id'])
        self.assertEqual('error', checkpoint_record['checkpoint_status'])

    def test_checkpoint_record_update_nonexistent(self):
        self.assertRaises(exception.CheckpointRecordNotFound,
                          db.checkpoint_record_update,
                          self.ctxt, 42, {})



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_versionedobjects import fields

from karbor import objects


def fake_db_operation_log(**updates):
    db_operation_log = {
        "id": "36ea41b2-c358-48a7-9117-70cb7617410a",
        "project_id": "586cc6ce-e286-40bd-b2b5-dd32694d9944",
        "scheduled_operation_id": "2220f8b1-975d-4621-a872-fa9afb43cb6c",
        "state": "failed",
        "error": "Could not access bank",
        "entries": "[entries:{'timestamp': '2015-08-27T09:50:51-05:00',"
                   "'message': 'Doing things'}]"
    }
    for name, field in objects.OperationLog.fields.items():
        if name in db_operation_log:
            continue
        if field.nullable:
            db_operation_log[name] = None
        elif field.default != fields.UnspecifiedDefault:
            db_operation_log[name] = field.default
        else:
            raise Exception('db_operation_log needs help with %s.' % name)

    if updates:
        db_operation_log.update(updates)

    return db_operation_log



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import timeutils
from oslo_versionedobjects import fields

from karbor import context
from karbor.objects import base as obj_base
from karbor.tests import base


class BaseObjectsTestCase(base.TestCase):
    def setUp(self):
        super(BaseObjectsTestCase, self).setUp()
        self.user_id = 'fake-user'
        self.project_id = 'fake-project'
        self.context = context.RequestContext(self.user_id, self.project_id,
                                              is_admin=False)
        # We only test local right now.
        self.assertIsNone(obj_base.KarborObject.indirection_api)

    @staticmethod
    def _compare(test, db, obj):
        for field, value in db.items():
            if not hasattr(obj, field):
                continue

            if (isinstance(obj.fields[field], fields.DateTimeField) and
               db[field]):
                test.assertEqual(db[field],
                                 timeutils.normalize_time(obj[field]))
            elif isinstance(obj[field], obj_base.ObjectListBase):
                test.assertEqual(db[field], obj[field].objects)
            else:
                test.assertEqual(db[field], obj[field])



# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from karbor.resource import Resource
from karbor.services.protection.protectable_plugin import ProtectablePlugin
from karbor.services.protection.protectable_registry import ProtectableRegistry

from karbor.tests import base

_FAKE_TYPE = "Karbor::Test::Fake"


class _FakeProtectablePlugin(ProtectablePlugin):
    def __init__(self, cntx):
        super(_FakeProtectablePlugin, self).__init__(cntx)
        self.graph = {}

    def instance(self, cntx):
        new = self.__class__(cntx)
        new.graph = self.graph
        return new

    def get_resource_type(self):
        return _FAKE_TYPE

    def get_parent_resource_types(self):
        return _FAKE_TYPE

    def list_resources(self, context):
        return self.graph.values()

    def show_resource(self, context, resource_id):
        return [Resource(type=_FAKE_TYPE,
                         id=resource.id,
                         name=resource.name)
                for resource in self.graph
                if resource.id == resource_id]

    def get_dependent_resources(self, context, parent_resource):
        return self.graph[parent_resource]


class ProtectableRegistryTest(base.TestCase):
    def setUp(self):
        super(ProtectableRegistryTest, self).setUp()
        self.protectable_registry = ProtectableRegistry()
        self._fake_plugin = _FakeProtectablePlugin(None)
        self.protectable_registry.register_plugin(self._fake_plugin)

    def test_graph_building(self):
        A = Resource(_FAKE_TYPE, "A", 'nameA')
        B = Resource(_FAKE_TYPE, "B", 'nameB')
        C = Resource(_FAKE_TYPE, "C", 'nameC')
        test_matrix = (
            (
                {A: [B],
                 B: [C],
                 C: []},
                (A, C)
            ),
            (
                {A: [C],
                 B: [C],
                 C: []},
                (A, C)
            ),
        )

        for g, resources in test_matrix:
            self._fake_plugin.graph = g
            result_graph = self.protectable_registry.build_graph(None,
                                                                 resources)
            self.assert_graph(result_graph, g)
            self.protectable_registry._protectable_map = {}

    def assert_graph(self, g, g_dict):
        for item in g:
            expected = set(g_dict[item.value])
            found = set(child.value for child in item.child_nodes)
            self.assertEqual(expected, found)
            self.assert_graph(item.child_nodes, g_dict)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from oslo_config import cfg

from karbor.common import karbor_keystone_plugin
from karbor import context
from karbor.services.operationengine import karbor_client
from karbor.tests import base


class KarborClientTest(base.TestCase):

    @mock.patch.object(karbor_keystone_plugin.KarborKeystonePlugin,
                       'get_service_endpoint')
    def test_create_client(self, get_service_endpoint):
        ctx = context.get_admin_context()
        ctx.project_id = '123'

        cfg.CONF.set_default('version', '1', 'karbor_client')

        karbor_url = "http://127.0.0.1:9090"
        sc = karbor_client.create(ctx, endpoint=karbor_url)
        self.assertEqual(karbor_url, sc.http_client.endpoint)

        karbor_url = "http://127.0.0.1:9090/$(project_id)s"
        get_service_endpoint.return_value = karbor_url
        endpoint = karbor_url.replace("$(project_id)s", ctx.project_id)
        sc = karbor_client.create(ctx)
        self.assertEqual(endpoint, sc.http_client.endpoint)
